# Support functionals

# fit initial interval censor survival model 
#----------------------------------------------------------------------------------
initial.surv = function (Time, d, W, WintF.vl, WintF.sl, id, times, method, parameterization, 
                         long = NULL, long.deriv=NULL,extra = NULL, LongFormat){
  
  old <- options(warn = (-1))
  on.exit(options(old))
  if (!is.null(long)) {
    long.id <- tapply(long, id, tail, 1) #long is the estimation 
    if (parameterization == "value") 
      longD.id <- NULL
  }
  idT <- extra$ii
  
  WW <- if (!LongFormat) {
    cbind(W, long.id, longD.id)
  }
  
  if (method %in% c("weibull-PH-GH", "piecewise-PH-GH")) {
    if (!LongFormat) {
      DD <- data.frame(id = id, Time = Time[id], d = d[id], times = times)
      if (!is.null(long)) {
        DD$long <- long * WintF.vl[id, , drop = FALSE]
        k <- ncol(DD$long)
      }
      if (!is.null(long.deriv)) {
        DD$longD <- long.deriv * WintF.sl[id, , drop = FALSE]
        l <- ncol(DD$longD)
      }
      dW <- as.data.frame(W[id, , drop = FALSE], row.names = row.names(DD))
      
      if (ncol(dW)) {
        names(dW) <- paste("W", seq_along(dW), sep = "")
        DD <- cbind(DD, dW)
      }
    }
    
    if (!LongFormat) {
      DD$start <- DD$times
      DD$stop <- unlist(lapply(split(DD[c("id", "start", "Time")], DD$id), function(d) c(d$start[-1], d$Time[1])))
      DD$event <- ave(DD$d, DD$id, FUN = function(x) {
        if (length(x) == 1) {
          x
        }
        else {
          x[seq(length(x) - 1)] <- 0
          x
        }
      })
    }
    
    baseCovs <- if (ncol(dW)) {
      paste("+", paste(names(dW), collapse = " + "))
    }
    else NULL
    
    form <- if (!LongFormat) {
      paste("Surv(start, stop, event) ~", "long", baseCovs)
    }
    
    if (!is.null(DD$strata)) 
      form <- paste(form, "+ strata(strata)")
    
    form <- as.formula(form)
    cph <- coxph(form, data = DD)  # fit the model
    coefs <- cph$coefficients
    out <- list(alpha = coefs[1:k], gammas = coefs[-(1:k)])
    
    # fit the baseline hazard function
    if (method == "weibull-PH-GH") {
      dat <- data.frame(Time = Time, d = d)
      init.fit <- survreg(Surv(Time, d) ~ WW, data = dat)
      coefs <- -init.fit$coef/init.fit$scale
      out$gammas <- c(coefs[1], out$gammas)
      out$sigma.t <- 1/init.fit$scale
    }
    
    if (method == "piecewise-PH-GH") {
      dat <- data.frame(Time = Time, d = d)
      cph. <- coxph(Surv(Time, d) ~ WW, data = dat, x = TRUE)
      init.fit <- piecewiseExp.ph(cph., knots = extra$control$knots)
      coefs <- init.fit$coef
      out$xi <- exp(coefs[grep("xi", names(coefs))])
    }
    
    return(out)
  }
} 


# estimating the parameters from longitudinal and survival jointly
#----------------------------------------------------------------------------------
piecewisePHGH.fit = function (x, y, id, initial.values, parameterization, derivForm, control){
  
  newenv <- new.env()
  
  environment(H.longPC) =parent.env(newenv)
  environment(gr.longPC) =parent.env(newenv)
  environment(gr.survPC) =parent.env(newenv)
  environment(opt.survPC) =parent.env(newenv)
  environment(LogLik.piecewiseGH) = parent.env(newenv)
  environment(Score.piecewiseGH) = parent.env(newenv)
  
  
  # survival part
  logT <- as.vector(y$logT)
  d <- as.vector(y$d)
  ind.D <- y$ind.D  # interval indicate
  
  # longitudinal part
  y <- as.vector(y$y)
  X <- x$X
  Xtime <- x$Xtime # baseline measure with survival time
  Xs <- x$Xs 
  
  Z <- as.matrix(x$Z)
  Ztime <- as.matrix(x$Ztime)
  Zs <- as.matrix(x$Zs)
  
  WW <- x$W
  WintF.vl <- x$WintF.vl
  WintF.sl <- x$WintF.sl
  Ws.intF.vl <- x$Ws.intF.vl
  Ws.intF.sl <- x$Ws.intF.sl
  
  X <- dropAttr(X)  
  Z <- dropAttr(Z)
  WW <-dropAttr(WW)
  WintF.vl <- dropAttr(WintF.vl)
  WintF.sl <- dropAttr(WintF.sl)
  Ws.intF.vl <- dropAttr(Ws.intF.vl)
  Ws.intF.sl <- dropAttr(Ws.intF.sl)
  
  if (parameterization == "value") {
    Xtime <- dropAttr(Xtime)
    Ztime <- dropAttr(Ztime)
    Xs <- dropAttr(Xs)
    Zs <- dropAttr(Zs)
  }
  
  indFixed <- derivForm$indFixed
  indRandom <- derivForm$indRandom
  
  ncx <- ncol(X)
  ncz <- ncol(Z)
  ncww <- if (!is.null(WW)) 
    ncol(WW)
  else 0
  
  n <- length(logT) # n number of subject
  N <- length(y) # N number of observation
  ni <- as.vector(tapply(id, id, length)) # number of visit for each subject
  
  Q <- x$Q # base hazard knots
  XtX <- crossprod(as.matrix(X))
  ZtZ <- lapply(split(Z, id), function(x) crossprod(matrix(x, ncol = ncz)))
  names(ZtZ) <- NULL
  ZtZ <- matrix(unlist(ZtZ), n, ncz * ncz, TRUE)
  outer.Ztime <- lapply(1:n, function(x) Ztime[x, ] %o% Ztime[x, ])
  
  GH <- gauher(control$GHk)
  b <- as.matrix(expand.grid(rep(list(GH$x), ncz)))
  k <- nrow(b)
  wGH <- as.matrix(expand.grid(rep(list(GH$w), ncz)))
  wGH <- 2^(ncz/2) * apply(wGH, 1, prod) * exp(rowSums(b * b))
  
  if (control$typeGH == "simple") {
    b <- sqrt(2) * t(control$inv.chol.VC %*% t(b))
    wGH <- wGH * control$det.inv.chol.VC
  }
  
  dimnames(b) <- NULL
  b2 <- if (ncz == 1) 
    b * b
  else t(apply(b, 1, function(x) x %o% x))
  Ztb <- Z %*% t(b)
  
  if (parameterization =="value") {
    Ztime.b <- as.matrix(Ztime) %*% t(b)
    Zsb <- as.matrix(Zs) %*% t(b)
  }
  
  nk <- control$GKk
  id.GK <- x$id.GK
  st <- x$st
  ind.K <- rep(unlist(lapply(ind.D, seq_len)), each = nk)
  wk <- unlist(lapply(ind.D, function(n) rep(x$wk, n)))
  wkP <- wk * rep(x$P, each = nk)
  
  betas <- as.vector(initial.values$betas)
  sigma <- initial.values$sigma
  gammas <- if (!is.null(WW)) 
    as.vector(initial.values$gammas)
  else NULL
  alpha <- as.vector(initial.values$alpha)
  Dalpha <- as.vector(initial.values$Dalpha)
  xi <- initial.values$xi
  D <- initial.values$D
  diag.D <- !is.matrix(D)
  
  if (!diag.D) 
    dimnames(D) <- NULL
  else names(D) <- NULL
  
  old <- options(warn = (-1))
  on.exit(options(old))
  iter <- control$iter.EM
  Y.mat <- matrix(0, iter + 1, ncx + 1)
  T.mat <- matrix(0, iter + 1, ncww + length(xi) + ncol(WintF.vl))
  B.mat <- if (diag.D) 
    matrix(0, iter + 1, ncz)
  else matrix(0, iter + 1, ncz * ncz)
  lgLik <- numeric(iter + 1)
  conv <- TRUE
  
  # interation
  for (it in 1:iter) {
    Y.mat[it, ] <- c(betas, sigma)
    T.mat[it, ] <- c(gammas, alpha, log(xi))
    B.mat[it, ] <- D
    eta.yx <- as.vector(X %*% betas)
    # estimate y longitudinal
    eta.tw <- if (!is.null(WW)) 
      as.vector(WW %*% gammas) # esitmate survival exp w.o longitudinal covariate
    else 0
    if (parameterization %in% c("value", "both")) {
      Y <- as.vector(Xtime %*% betas) + Ztime.b
      Ys <- as.vector(Xs %*% betas) + Zsb
      eta.t <- eta.tw + c(WintF.vl %*% alpha) * Y
      eta.s <- c(Ws.intF.vl %*% alpha) * Ys
    }
    
    mu.y <- eta.yx + Ztb # mean y
    logNorm <- dnorm(y, mu.y, sigma, TRUE)
    log.p.yb <- rowsum(logNorm, id, reorder = FALSE)
    
    dimnames(log.p.yb) <- NULL
    log.hazard <- log(xi[ind.D]) + eta.t
    log.survival <- -exp(eta.tw) * rowsum(xi[ind.K] * wkP *exp(eta.s), id.GK, reorder = FALSE)
    
    dimnames(log.survival) <- NULL
    log.p.tb <- d * log.hazard + log.survival
    log.p.b <- if (control$typeGH == "simple") {
      rep(dmvnorm(b, rep(0, ncz), D, TRUE), each = n)
    }
    
    p.ytb <- exp(log.p.yb + log.p.tb + log.p.b)
    if (control$typeGH != "simple") 
      p.ytb <- p.ytb * VCdets
    p.yt <- c(p.ytb %*% wGH)
    p.byt <- p.ytb/p.yt
    post.b <- if (control$typeGH == "simple") {
      p.byt %*% (b * wGH)
    }
    
    post.vb <- if (control$typeGH == "simple") {
      if (ncz == 1) {
        c(p.byt %*% (b2 * wGH)) - c(post.b * post.b)
      }
      else {
        (p.byt %*% (b2 * wGH)) - t(apply(post.b, 1, function(x) x %o% 
                                           x))
      }
    }
    
    log.p.yt <- log(p.yt)
    lgLik[it] <- sum(log.p.yt[is.finite(log.p.yt)], na.rm = TRUE)
    if (control$verbose) {
      cat("\n\niter:", it, "\n")
      cat("log-likelihood:", lgLik[it], "\n")
      cat("betas:", round(betas, 4), "\n")
      cat("sigma:", round(sigma, 4), "\n")
      if (!is.null(WW)) 
        cat("gammas:", round(gammas, 4), "\n")
      if (parameterization %in% c("value", "both")) 
        cat("alpha:", round(alpha, 4), "\n")
      if (parameterization %in% c("slope", "both")) 
        cat("alphaD:", round(Dalpha, 4), "\n")
      cat("xi:", round(xi, 4), "\n")
      cat("D:", if (!diag.D) 
        round(D[lower.tri(D, TRUE)], 4)
        else round(D, 4), "\n")
    }
    if (it > 5 && lgLik[it] > lgLik[it - 1]) {
      thets1 <- c(Y.mat[it - 1, ], T.mat[it - 1, ], B.mat[it -1, ])
      thets2 <- c(Y.mat[it, ], T.mat[it, ], B.mat[it, ])
      check1 <- max(abs(thets2 - thets1)/(abs(thets1) + 
                                            control$tol1)) < control$tol2
      check2 <- (lgLik[it] - lgLik[it - 1]) < control$tol3 * 
        (abs(lgLik[it - 1]) + control$tol3)
      if (check1 || check2) {
        conv <- FALSE
        if (control$verbose) 
          cat("\n\nconverged!\n")
        break
      }
    }
    if (iter == 0) 
      break
    
    Zb <- rowSums(Z * post.b[id, ], na.rm = TRUE)
    mu <- y - eta.yx
    tr.tZZvarb <- sum(ZtZ * post.vb, na.rm = TRUE)
    sigman <- sqrt(c(crossprod(mu, mu - 2 * Zb) + crossprod(Zb) + tr.tZZvarb)/N)
    Dn <- if (control$typeGH == "simple") {
      matrix(colMeans(p.byt %*% (b2 * wGH), na.rm = TRUE), 
             ncz, ncz)
    }
    Dn <- if (diag.D) 
      diag(Dn)
    else 0.5 * (Dn + t(Dn))
    
    Hbetas <- nearPD(H.longPC(betas))
    scbetas <- gr.longPC(betas)
    betasn <- betas - c(solve(Hbetas, scbetas))
    list.thetas <- list(gammas = gammas, alpha = alpha, Dalpha = Dalpha, log.xi = log(xi))
    list.thetas <- list.thetas[!sapply(list.thetas, is.null)]
    thetas <- unlist(as.relistable(list.thetas))
    gr.survPC(thetas)
    optz.surv <- optim(thetas, opt.survPC, gr.survPC, method = "BFGS", 
                       control = list(maxit = if (it < 5) 20 else 5, parscale = if (it < 5) rep(0.01, length(thetas)) else rep(0.1, length(thetas))))
    thetasn <- relist(optz.surv$par, skeleton = list.thetas)
    betas <- betasn
    sigma <- sigman
    D <- Dn
    gammas <- thetasn$gammas
    alpha <- thetasn$alpha
    Dalpha <- thetasn$Dalpha
    xi <- exp(thetasn$log.xi)
  }
  
  #---------------------------------------------------
  
  # use newton method if not converge
  list.thetas <- list(betas = betas, log.sigma = log(sigma), 
                      gammas = gammas, alpha = alpha, Dalpha = Dalpha, log.xi = log(xi), 
                      D = if (diag.D) log(D) else chol.transf(D))
  list.thetas <- list.thetas[!sapply(list.thetas, is.null)]
  thetas <- unlist(as.relistable(list.thetas))
  lgLik <- -LogLik.piecewiseGH(thetas)
  
  if (conv && !control$only.EM) {
    if (is.null(control$parscale)) 
      control$parscale <- rep(0.01, length(thetas))
    if (control$verbose) 
      cat("\n\nquasi-Newton iterations start.\n\n")
    out <- if (control$optimizer == "optim") {
      optim(thetas, LogLik.piecewiseGH, Score.piecewiseGH, 
            method = "BFGS", control = list(maxit = control$iter.qN, 
                                            parscale = control$parscale, trace = 10 * control$verbose))
    }
    if ((conv <- out$convergence) == 0 || -out[[2]] > lgLik) {
      lgLik <- -out[[2]]
      thetas <- relist(out$par, skeleton = list.thetas)
      betas <- thetas$betas
      sigma <- exp(thetas$log.sigma)
      gammas <- thetas$gammas
      alpha <- thetas$alpha
      Dalpha <- thetas$Dalpha
      xi <- exp(thetas$log.xi)
      D <- thetas$D
      D <- if (diag.D) 
        exp(D)
      else chol.transf(D)
      it <- it + if (control$optimizer == "optim") 
        out$counts[1]
      else out$iterations
      eta.yx <- as.vector(X %*% betas)
      eta.tw <- if (!is.null(WW)) 
        as.vector(WW %*% gammas)
      else 0
      if (parameterization %in% c("value")) {
        Y <- as.vector(Xtime %*% betas) + Ztime.b
        Ys <- as.vector(Xs %*% betas) + Zsb
        eta.t <- eta.tw + c(WintF.vl %*% alpha) * Y
        eta.s <- c(Ws.intF.vl %*% alpha) * Ys
      }
      
      exp.eta.tw <- exp(eta.tw)
      mu.y <- eta.yx + Ztb
      logNorm <- dnorm(y, mu.y, sigma, TRUE)
      log.p.yb <- rowsum(logNorm, id)
      log.hazard <- log(xi[ind.D]) + eta.t
      log.survival <- -exp(eta.tw) * rowsum(xi[ind.K] * 
                                              wkP * exp(eta.s), id.GK, reorder = FALSE)
      dimnames(log.survival) <- NULL
      log.p.tb <- d * log.hazard + log.survival
      log.p.b <- if (control$typeGH == "simple") {
        rep(dmvnorm(b, rep(0, ncz), D, TRUE), each = n)
      }
      else {
        matrix(dmvnorm(do.call(rbind, lis.b), rep(0, 
                                                  ncz), D, TRUE), n, k, byrow = TRUE)
      }
      
      
      p.ytb <- exp(log.p.yb + log.p.tb + log.p.b)
      if (control$typeGH != "simple") 
        p.ytb <- p.ytb * VCdets
      p.yt <- c(p.ytb %*% wGH)
      p.byt <- p.ytb/p.yt
      post.b <- if (control$typeGH == "simple") {
        p.byt %*% (b * wGH)
      }
      else {
        sapply(seq_len(ncz), function(i) (p.byt * t(sapply(lis.b, 
                                                           "[", seq_len(k), i))) %*% wGH)
      }
      post.vb <- if (control$typeGH == "simple") {
        if (ncz == 1) {
          c(p.byt %*% (b2 * wGH)) - c(post.b * post.b)
        }
        else {
          (p.byt %*% (b2 * wGH)) - t(apply(post.b, 1, 
                                           function(x) x %o% x))
        }
      }
      else {
        dd <- sapply(seq_len(ncz^2), function(i) (p.byt * 
                                                    t(sapply(lis.b2, "[", seq_len(k), i))) %*% 
                       wGH)
        bb <- apply(post.b, 1, function(x) x %o% x)
        dd - if (ncz == 1) 
          c(bb)
        else t(bb)
      }
      Zb <- if (ncz == 1) 
        post.b[id]
      else rowSums(Z * post.b[id, ], na.rm = TRUE)
    }
  }
  
  #----------------------------------------------------
  
  Score <- Score.piecewiseGH(unlist(thetas))
  if (control$verbose) 
    cat("\ncalculating Hessian...\n")
  Hessian <- if (control$numeriDeriv == "fd") {
    fd.vec(unlist(thetas), Score.piecewiseGH, eps = control$eps.Hes)
  }
  names(betas) <- names(initial.values$betas)
  if (!diag.D) 
    dimnames(D) <- dimnames(initial.values$D)
  if (!is.null(WW)) 
    names(gammas) <- colnames(x$W)
  nm.alph <- colnames(x$WintF.vl)
  nm.alph <- if (!is.null(nm.alph)) {
    if (nm.alph[1] == "(Intercept)") 
      c("", nm.alph[-1])
    else nm.alph
  }
  else {
    "alpha"
  }
  nm.Dalph <- colnames(x$WintF.sl)
  nm.Dalph <- if (!is.null(nm.Dalph)) {
    if (nm.Dalph[1] == "(Intercept)") 
      c("", nm.Dalph[-1])
    else nm.Dalph
  }
  else {
    "alpha.s"
  }
  
  gg <- switch(parameterization, value = nm.alph, slope = nm.Dalph, 
               both = c(nm.alph, nm.Dalph))
  if (parameterization %in% c("value", "both")) 
    names(alpha) <- nm.alph
  if (parameterization %in% c("slope", "both")) 
    names(Dalpha) <- nm.Dalph
  names(xi) <- paste("xi.", seq_len(Q), sep = "")
  nams <- c(paste("Y.", c(names(betas), "sigma"), sep = ""), 
            paste("T.", c(names(gammas), gg, names(xi)), sep = ""), 
            paste("B.", if (!diag.D) paste("D", seq(1, ncz * (ncz +  1)/2), sep = "") else names(D), sep = ""))
  dimnames(Hessian) <- list(nams, nams)
  colnames(post.b) <- colnames(x$Z)
  out = list(coefficients = list(betas = betas, sigma = sigma, gammas = gammas, 
                                 alpha = alpha, Dalpha = Dalpha, xi = xi, D = as.matrix(D)), 
             Score = Score, Hessian = Hessian, logLik = lgLik, EB = list(post.b = post.b, post.vb = post.vb, 
                                                                         Zb = if (iter == 0) rowSums(Z * post.b[id, ], na.rm = TRUE) else Zb, Ztimeb = if (parameterization %in% c("value", "both")) rowSums(Ztime * post.b) else NULL, 
                                                                         Ztimeb.deriv = if (parameterization %in% c("slope","both")) {
                                                                           if (indRandom) rowSums(Ztime.deriv * post.b[,indRandom, drop = FALSE]) else rep(0, nrow(Ztime.deriv))
                                                                         } else NULL), iters = it, convergence = conv, n = n, N = N, ni = ni, d = d, id = id)
  
  return(out)
}
#----------------------------------------------------------------------------------

weibullPHGH.fit = function (x, y, id, initial.values, scaleWB, parameterization, derivForm, control) {
  logT <- as.vector(y$logT)
  d <- as.vector(y$d)
  y <- as.vector(y$y)
  X <- x$X
  Xtime <- x$Xtime
  Xs <- x$Xs
  Xtime.deriv <- x$Xtime.deriv
  Xs.deriv <- x$Xs.deriv
  Z <- x$Z
  Ztime <- x$Ztime
  Zs <- x$Zs
  Ztime.deriv <- x$Ztime.deriv
  Zs.deriv <- x$Zs.deriv
  W1 <- x$W
  WW <- if (is.null(W1)) 
    as.matrix(rep(1, length(logT)))
  else cbind(1, W1)
  WintF.vl <- x$WintF.vl
  WintF.sl <- x$WintF.sl
  Ws.intF.vl <- x$Ws.intF.vl
  Ws.intF.sl <- x$Ws.intF.sl
  X <- dropAttr(X)
  Z <- dropAttr(Z)
  WW <- dropAttr(WW)
  WintF.vl <- dropAttr(WintF.vl)
  WintF.sl <- dropAttr(WintF.sl)
  Ws.intF.vl <- dropAttr(Ws.intF.vl)
  Ws.intF.sl <- dropAttr(Ws.intF.sl)
  if (parameterization == "value") {
    Xtime <- dropAttr(Xtime)
    Ztime <- dropAttr(Ztime)
    Xs <- dropAttr(Xs)
    Zs <- dropAttr(Zs)
  }
  indFixed <- derivForm$indFixed
  indRandom <- derivForm$indRandom
  ncx <- ncol(X)
  ncz <- ncol(Z)
  ncww <- ncol(WW)
  n <- length(logT)
  N <- length(y)
  ni <- as.vector(tapply(id, id, length))
  XtX <- crossprod(X)
  ZtZ <- lapply(split(Z, id), function(x) crossprod(matrix(x, ncol = ncz)))
  names(ZtZ) <- NULL
  ZtZ <- matrix(unlist(ZtZ), n, ncz * ncz, TRUE)
  outer.Ztime <- lapply(1:n, function(x) Ztime[x, ] %o% Ztime[x, ])
  GH <- gauher(control$GHk)
  b <- as.matrix(expand.grid(rep(list(GH$x), ncz)))
  k <- nrow(b)
  wGH <- as.matrix(expand.grid(rep(list(GH$w), ncz)))
  wGH <- 2^(ncz/2) * apply(wGH, 1, prod) * exp(rowSums(b * b))
  
  if (control$typeGH == "simple") {
    b <- sqrt(2) * t(control$inv.chol.VC %*% t(b))
    wGH <- wGH * control$det.inv.chol.VC
  }
  else {
    b <- sqrt(2) * b
    VCdets <- control$det.inv.chol.VCs
  }
  dimnames(b) <- NULL
  b2 <- if (ncz == 1) 
    b * b
  else t(apply(b, 1, function(x) x %o% x))
  
  Ztb <- Z %*% t(b)
  if (parameterization =="value") {
    Ztime.b <- Ztime %*% t(b)
    Zsb <- Zs %*% t(b)
  }
  
  st <- x$st
  log.st <- log(st)
  wk <- rep(x$wk, length(logT))
  P <- as.vector(x$P)
  id.GK <- rep(seq_along(logT), each = control$GKk)
  
  betas <- as.vector(initial.values$betas)
  sigma <- initial.values$sigma
  gammas <- as.vector(initial.values$gammas)
  alpha <- as.vector(initial.values$alpha)
  Dalpha <- as.vector(initial.values$Dalpha)
  sigma.t <- if (is.null(scaleWB)) 
    initial.values$sigma.t
  else scaleWB
  D <- initial.values$D
  diag.D <- !is.matrix(D)
  if (!diag.D) 
    dimnames(D) <- NULL
  else names(D) <- NULL
  
  newenv <- new.env()
  environment(opt.survWB) <- environment(gr.survWB) <- parent.env(newenv)
  environment(opt.longWB) <- environment(gr.longWB) <- environment(H.longWB) <- parent.env(newenv)
  environment(LogLik.weibullGH) <- environment(Score.weibullGH) <- parent.env(newenv)
  
  old <- options(warn = (-1))
  on.exit(options(old))
  iter <- control$iter.EM
  Y.mat <- matrix(0, iter + 1, ncx + 1)
  T.mat <- matrix(0, iter + 1, switch(parameterization, value = ncww + 
                                        1 + ncol(WintF.vl), slope = ncww + 1 + ncol(WintF.sl), 
                                      both = ncww + 1 + ncol(WintF.vl) + ncol(WintF.sl)))
  B.mat <- if (diag.D) 
    matrix(0, iter + 1, ncz)
  else matrix(0, iter + 1, ncz * ncz)
  
  lgLik <- numeric(iter + 1)
  conv <- TRUE
  for (it in 1:iter) {
    Y.mat[it, ] <- c(betas, sigma)
    
    T.mat[it, ] <- switch(parameterization, value = c(gammas, alpha, sigma.t), slope = c(gammas, Dalpha, sigma.t), 
                          both = c(gammas, alpha, Dalpha, sigma.t))
    B.mat[it, ] <- D
    eta.yx <- as.vector(X %*% betas)
    eta.tw <- as.vector(WW %*% gammas)
    if (parameterization =="value") {
      Y <- as.vector(Xtime %*% betas) + Ztime.b
      Ys <- as.vector(Xs %*% betas) + Zsb
      eta.t <- eta.tw + c(WintF.vl %*% alpha) * Y
      eta.s <- c(Ws.intF.vl %*% alpha) * Ys
    }
    mu.y <- eta.yx + Ztb
    logNorm <- dnorm(y, mu.y, sigma, TRUE)
    log.p.yb <- rowsum(logNorm, id, reorder = FALSE)
    dimnames(log.p.yb) <- NULL
    log.hazard <- log(sigma.t) + (sigma.t - 1) * logT + eta.t
    log.survival <- -exp(eta.tw) * P * rowsum(wk * exp(log(sigma.t) +(sigma.t - 1) * log.st + eta.s), id.GK, reorder = FALSE)
    dimnames(log.survival) <- NULL
    log.p.tb <- d * log.hazard + log.survival
    log.p.b <- if (control$typeGH == "simple") {
      rep(dmvnorm(b, rep(0, ncz), D, TRUE), each = n)
    }
    else {
      matrix(dmvnorm(do.call(rbind, lis.b), rep(0, ncz), 
                     D, TRUE), n, k, byrow = TRUE)
    }
    p.ytb <- exp(log.p.yb + log.p.tb + log.p.b)
    if (control$typeGH != "simple") 
      p.ytb <- p.ytb * VCdets
    p.yt <- c(p.ytb %*% wGH)
    p.byt <- p.ytb/p.yt
    post.b <- if (control$typeGH == "simple") {
      p.byt %*% (b * wGH)
    }
    else {
      sapply(seq_len(ncz), function(i) (p.byt * t(sapply(lis.b, 
                                                         "[", seq_len(k), i))) %*% wGH)
    }
    post.vb <- if (control$typeGH == "simple") {
      if (ncz == 1) {
        c(p.byt %*% (b2 * wGH)) - c(post.b * post.b)
      }
      else {
        (p.byt %*% (b2 * wGH)) - t(apply(post.b, 1, function(x) x %o% 
                                           x))
      }
    }
    else {
      dd <- sapply(seq_len(ncz^2), function(i) (p.byt * 
                                                  t(sapply(lis.b2, "[", seq_len(k), i))) %*% wGH)
      bb <- apply(post.b, 1, function(x) x %o% x)
      dd - if (ncz == 1) 
        c(bb)
      else t(bb)
    }
    
    log.p.yt <- log(p.yt)
    lgLik[it] <- sum(log.p.yt[is.finite(log.p.yt)], na.rm = TRUE)
    if (control$verbose) {
      cat("\n\niter:", it, "\n")
      cat("log-likelihood:", lgLik[it], "\n")
      cat("betas:", round(betas, 4), "\n")
      cat("sigma:", round(sigma, 4), "\n")
      cat("gammas:", round(gammas, 4), "\n")
      if (parameterization %in% c("value", "both")) 
        cat("alpha:", round(alpha, 4), "\n")
      if (parameterization %in% c("slope", "both")) 
        cat("alphaD:", round(Dalpha, 4), "\n")
      cat("sigma.t:", round(sigma.t, 4), "\n")
      cat("D:", if (!diag.D) 
        round(D[lower.tri(D, TRUE)], 4)
        else round(D, 4), "\n")
    }
    if (it > 5 && lgLik[it] > lgLik[it - 1]) {
      thets1 <- c(Y.mat[it - 1, ], T.mat[it - 1, ], B.mat[it - 
                                                            1, ])
      thets2 <- c(Y.mat[it, ], T.mat[it, ], B.mat[it, ])
      check1 <- max(abs(thets2 - thets1)/(abs(thets1) + 
                                            control$tol1)) < control$tol2
      check2 <- (lgLik[it] - lgLik[it - 1]) < control$tol3 * 
        (abs(lgLik[it - 1]) + control$tol3)
      if (check1 || check2) {
        conv <- FALSE
        if (control$verbose) 
          cat("\n\nconverged!\ncalculating Hessian...\n")
        break
      }
    }
    if (iter == 0) 
      break
    Zb <- rowSums(Z * post.b[id, ], na.rm = TRUE)
    mu <- y - eta.yx
    tr.tZZvarb <- sum(ZtZ * post.vb, na.rm = TRUE)
    sigman <- sqrt(c(crossprod(mu, mu - 2 * Zb) + crossprod(Zb) + tr.tZZvarb)/N)
    Dn <- if (control$typeGH == "simple") {
      matrix(colMeans(p.byt %*% (b2 * wGH), na.rm = TRUE), 
             ncz, ncz)
    }
    else {
      matrix(colMeans(dd, na.rm = TRUE), ncz, ncz)
    }
    Dn <- if (diag.D) 
      diag(Dn)
    else 0.5 * (Dn + t(Dn))
    Hbetas <- nearPD(H.longWB(betas))
    scbetas <- gr.longWB(betas)
    betasn <- betas - c(solve(Hbetas, scbetas))
    list.thetas <- list(gammas = gammas, alpha = alpha, Dalpha = Dalpha, 
                        log.sigma.t = log(sigma.t))
    if (!is.null(scaleWB)) 
      list.thetas$log.sigma.t <- NULL
    list.thetas <- list.thetas[!sapply(list.thetas, is.null)]
    thetas <- unlist(as.relistable(list.thetas))
    optz.surv <- optim(thetas, opt.survWB, gr.survWB, method = "BFGS", 
                       control = list(maxit = if (it < 5) 20 else 5, parscale = if (it < 
                                                                                    5) rep(0.01, length(thetas)) else rep(0.1, length(thetas))))
    thetasn <- relist(optz.surv$par, skeleton = list.thetas)
    betas <- betasn
    sigma <- sigman
    D <- Dn
    gammas <- thetasn$gammas
    alpha <- thetasn$alpha
    Dalpha <- thetasn$Dalpha
    sigma.t <- if (is.null(scaleWB)) 
      exp(thetasn$log.sigma.t)
    else scaleWB
  }
  list.thetas <- list(betas = betas, log.sigma = log(sigma), 
                      gammas = gammas, alpha = alpha, Dalpha = Dalpha, log.sigma.t = log(sigma.t), 
                      D = if (diag.D) log(D) else chol.transf(D))
  if (!is.null(scaleWB)) 
    list.thetas$log.sigma.t <- NULL
  list.thetas <- list.thetas[!sapply(list.thetas, is.null)]
  thetas <- unlist(as.relistable(list.thetas))
  lgLik <- -LogLik.weibullGH(thetas)
  
  #----------------------------------------
  if (conv && !control$only.EM) {
    if (is.null(control$parscale)) 
      control$parscale <- rep(0.01, length(thetas))
    if (control$verbose) 
      cat("\n\nquasi-Newton iterations start.\n\n")
    out <- if (control$optimizer == "optim") {
      optim(thetas, LogLik.weibullGH, Score.weibullGH, 
            method = "BFGS", control = list(maxit = control$iter.qN, 
                                            parscale = control$parscale, trace = 10 * control$verbose))
    }
    else {
      nlminb(thetas, LogLik.weibullGH, Score.weibullGH, 
             scale = control$parscale, control = list(iter.max = control$iter.qN, 
                                                      trace = 1 * control$verbose))
    }
    if ((conv <- out$convergence) == 0 || -out[[2]] > lgLik) {
      lgLik <- -out[[2]]
      thetas <- relist(out$par, skeleton = list.thetas)
      betas <- thetas$betas
      sigma <- exp(thetas$log.sigma)
      gammas <- thetas$gammas
      alpha <- thetas$alpha
      Dalpha <- thetas$Dalpha
      sigma.t <- if (is.null(scaleWB)) 
        exp(thetas$log.sigma.t)
      else scaleWB
      D <- thetas$D
      D <- if (diag.D) 
        exp(D)
      else chol.transf(D)
      it <- it + if (control$optimizer == "optim") 
        out$counts[1]
      else out$iterations
      eta.yx <- as.vector(X %*% betas)
      eta.tw <- as.vector(WW %*% gammas)
      if (parameterization %in% c("value", "both")) {
        Y <- as.vector(Xtime %*% betas) + Ztime.b
        Ys <- as.vector(Xs %*% betas) + Zsb
        eta.t <- eta.tw + c(WintF.vl %*% alpha) * Y
        eta.s <- c(Ws.intF.vl %*% alpha) * Ys
      }
      if (parameterization %in% c("slope", "both")) {
        Y.deriv <- as.vector(Xtime.deriv %*% betas[indFixed]) + 
          Ztime.b.deriv
        Ys.deriv <- as.vector(Xs.deriv %*% betas[indFixed]) + 
          Zsb.deriv
        eta.t <- if (parameterization == "both") 
          eta.t + c(WintF.sl %*% Dalpha) * Y.deriv
        else eta.tw + c(WintF.sl %*% Dalpha) * Y.deriv
        eta.s <- if (parameterization == "both") 
          eta.s + c(Ws.intF.sl %*% Dalpha) * Ys.deriv
        else c(Ws.intF.sl %*% Dalpha) * Ys.deriv
      }
      exp.eta.tw <- exp(eta.tw)
      mu.y <- eta.yx + Ztb
      logNorm <- dnorm(y, mu.y, sigma, TRUE)
      log.p.yb <- rowsum(logNorm, id)
      log.hazard <- log(sigma.t) + (sigma.t - 1) * logT + 
        eta.t
      log.survival <- -exp.eta.tw * P * rowsum(wk * exp(log(sigma.t) + 
                                                          (sigma.t - 1) * log.st + eta.s), id.GK, reorder = FALSE)
      dimnames(log.survival) <- NULL
      log.p.tb <- d * log.hazard + log.survival
      log.p.b <- if (control$typeGH == "simple") {
        rep(dmvnorm(b, rep(0, ncz), D, TRUE), each = n)
      }
      else {
        matrix(dmvnorm(do.call(rbind, lis.b), rep(0, 
                                                  ncz), D, TRUE), n, k, byrow = TRUE)
      }
      p.ytb <- exp(log.p.yb + log.p.tb + log.p.b)
      if (control$typeGH != "simple") 
        p.ytb <- p.ytb * VCdets
      p.yt <- c(p.ytb %*% wGH)
      p.byt <- p.ytb/p.yt
      post.b <- if (control$typeGH == "simple") {
        p.byt %*% (b * wGH)
      }
      else {
        sapply(seq_len(ncz), function(i) (p.byt * t(sapply(lis.b, 
                                                           "[", seq_len(k), i))) %*% wGH)
      }
      post.vb <- if (control$typeGH == "simple") {
        if (ncz == 1) {
          c(p.byt %*% (b2 * wGH)) - c(post.b * post.b)
        }
        else {
          (p.byt %*% (b2 * wGH)) - t(apply(post.b, 1, 
                                           function(x) x %o% x))
        }
      }
      else {
        dd <- sapply(seq_len(ncz^2), function(i) (p.byt * 
                                                    t(sapply(lis.b2, "[", seq_len(k), i))) %*% 
                       wGH)
        bb <- apply(post.b, 1, function(x) x %o% x)
        dd - if (ncz == 1) 
          c(bb)
        else t(bb)
      }
      Zb <- if (ncz == 1) 
        post.b[id]
      else rowSums(Z * post.b[id, ], na.rm = TRUE)
    }
  }
  
  
  Score <- Score.weibullGH(unlist(thetas))
  Hessian <- if (control$numeriDeriv == "fd") {
    fd.vec(unlist(thetas), Score.weibullGH, eps = control$eps.Hes)
  }
  else {
    cd.vec(unlist(thetas), Score.weibullGH, eps = control$eps.Hes)
  }
  names(betas) <- names(initial.values$betas)
  if (!diag.D) 
    dimnames(D) <- dimnames(initial.values$D)
  else names(D) <- names(initial.values$D)
  names(gammas) <- c("(Intercept)", colnames(W1))
  nm.alph <- colnames(x$WintF.vl)
  nm.alph <- if (!is.null(nm.alph)) {
    if (nm.alph[1] == "(Intercept)") 
      c("", nm.alph[-1])
    else nm.alph
  }
  else {
    "alpha"
  }
  nm.Dalph <- colnames(x$WintF.sl)
  nm.Dalph <- if (!is.null(nm.Dalph)) {
    if (nm.Dalph[1] == "(Intercept)") 
      c("", nm.Dalph[-1])
    else nm.Dalph
  }
  else {
    "alpha.s"
  }
  gg <- nm.alph 
  names(alpha) <- nm.alph
  nams <- c(paste("Y.", c(names(betas), "sigma"), sep = ""), 
            paste("T.", if (is.null(scaleWB)) c(names(gammas), gg, 
                                                "sigma.t") else c(names(gammas), gg), sep = ""), 
            paste("B.", if (!diag.D) paste("D", seq(1, ncz * (ncz + 
                                                                1)/2), sep = "") else names(D), sep = ""))
  dimnames(Hessian) <- list(nams, nams)
  colnames(post.b) <- colnames(x$Z)
  list(coefficients = list(betas = betas, sigma = sigma, gammas = gammas, 
                           alpha = alpha, Dalpha = Dalpha, sigma.t = sigma.t, D = as.matrix(D)), 
       Score = Score, Hessian = Hessian, logLik = lgLik, EB = list(post.b = post.b, 
                                                                   post.vb = post.vb, 
                                                                   Zb = if (iter == 0) rowSums(Z * post.b[id, ], na.rm = TRUE) else Zb, Ztimeb = if (parameterization %in% 
                                                                                                                                                     c("value", "both")) rowSums(Ztime * post.b) else NULL, 
                                                                   Ztimeb.deriv = if (parameterization %in% c("slope", "both")) {
                                                                     if (indRandom) rowSums(Ztime.deriv * post.b[, indRandom, drop = FALSE]) else rep(0, nrow(Ztime.deriv))
                                                                   } else NULL), iters = it, convergence = conv, n = n, 
       N = N, ni = ni, d = d, id = id, scaleWB = scaleWB)
}


gaussKronrod = function (k = 15) {
  sk <- c(-0.949107912342758, -0.741531185599394, -0.405845151377397, 
          0, 0.405845151377397, 0.741531185599394, 0.949107912342758, 
          -0.991455371120813, -0.864864423359769, -0.586087235467691, 
          -0.207784955007898, 0.207784955007898, 0.586087235467691, 
          0.864864423359769, 0.991455371120813)
  wk15 <- c(0.0630920926299786, 0.140653259715526, 0.190350578064785, 
            0.209482141084728, 0.190350578064785, 0.140653259715526, 
            0.0630920926299786, 0.0229353220105292, 0.10479001032225, 
            0.169004726639268, 0.204432940075299, 0.204432940075299, 
            0.169004726639268, 0.10479001032225, 0.0229353220105292)
  wk7 <- c(0.12948496616887, 0.279705391489277, 0.381830050505119, 
           0.417959183673469, 0.381830050505119, 0.279705391489277, 
           0.12948496616887)
  if (k == 7) 
    list(sk = sk[1:7], wk = wk7)
  else list(sk = sk, wk = wk15)
}

# piecewise proportional hazard
piecewiseExp.ph = function (coxObject, knots = NULL, length.knots = 6) {
  Time <- coxObject$y[, 1]
  d <- coxObject$y[, 2]
  n <- length(Time)
  if (is.null(knots)) {
    Q <- length.knots + 1
    knots <- unique(quantile(Time, seq(0, 1, len = Q + 1), 
                             names = FALSE)[-c(1, Q + 1)])
    knots <- knots + 1e-06
    if (max(knots) > max(Time)) 
      knots[which.max(knots)] <- max(Time) - 1e-06
  }
  knots <- c(0, sort(knots), max(Time) + 1)
  Q <- length(knots) - 1
  ind <- findInterval(Time, knots, rightmost.closed = TRUE)
  D <- matrix(0, n, Q)
  D[cbind(seq_along(ind), ind)] <- 1
  D <- c(D * d)
  Tiq <- outer(Time, knots, pmin)
  T <- c(Tiq[, 2:(Q + 1)] - Tiq[, 1:Q])
  X <- coxObject$x[rep(seq_len(n), Q), ]
  ND <- suppressWarnings(data.frame(Time = T, D = D, X, xi = gl(Q, 
                                                                n), check.names = FALSE)[T > 0, ])
  glm(D ~ . + offset(log(Time)) - Time - 1, family = poisson, 
      data = ND)
}

gauher = function (n) {
  m <- trunc((n + 1)/2)
  x <- w <- rep(-1, n)
  for (i in seq_len(m)) {
    z <- if (i == 1) {
      sqrt(2 * n + 1) - 1.85575 * (2 * n + 1)^(-0.16667)
    }
    else if (i == 2) {
      z - 1.14 * n^0.426/z
    }
    else if (i == 3) {
      1.86 * z - 0.86 * x[1]
    }
    else if (i == 4) {
      1.91 * z - 0.91 * x[2]
    }
    else {
      2 * z - x[i - 2]
    }
    for (its in seq_len(10)) {
      p1 <- 0.751125544464943
      p2 <- 0
      for (j in seq_len(n)) {
        p3 <- p2
        p2 <- p1
        p1 <- z * sqrt(2/j) * p2 - sqrt((j - 1)/j) * 
          p3
      }
      pp <- sqrt(2 * n) * p2
      z1 <- z
      z <- z1 - p1/pp
      if (abs(z - z1) <= 3e-14) 
        break
    }
    x[i] <- z
    x[n + 1 - i] <- -z
    w[i] <- 2/(pp * pp)
    w[n + 1 - i] <- w[i]
  }
  list(x = x, w = w)
}

dmvnorm = function (x, mu, Sigma, log = FALSE) {
  if (!is.matrix(x)) 
    x <- rbind(x)
  p <- length(mu)
  if (p == 1) {
    dnorm(x, mu, sqrt(Sigma), log = log)
  }
  else {
    t1 <- length(mu) == length(Sigma)
    t2 <- all(abs(Sigma[lower.tri(Sigma)]) < sqrt(.Machine$double.eps))
    if (t1 || t2) {
      if (!t1) 
        Sigma <- diag(Sigma)
      nx <- nrow(x)
      ff <- rowSums(dnorm(x, rep(mu, each = nx), sd = rep(sqrt(Sigma), 
                                                          each = nx), log = TRUE))
      if (log) 
        ff
      else exp(ff)
    }
    else {
      ed <- eigen(Sigma, symmetric = TRUE)
      ev <- ed$values
      evec <- ed$vectors
      if (!all(ev >= -1e-06 * abs(ev[1]))) 
        stop("'Sigma' is not positive definite")
      ss <- x - rep(mu, each = nrow(x))
      inv.Sigma <- evec %*% (t(evec)/ev)
      quad <- 0.5 * rowSums((ss %*% inv.Sigma) * ss)
      fact <- -0.5 * (p * log(2 * pi) + sum(log(ev)))
      if (log) 
        as.vector(fact - quad)
      else as.vector(exp(fact - quad))
    }
  }
}


nearPD = function (M, eig.tol = 1e-06, conv.tol = 1e-07, posd.tol = 1e-08,  maxits = 100) {
  if (!(is.numeric(M) && is.matrix(M) && identical(M, t(M)))) 
    stop("Input matrix M must be square and symmetric.\n")
  inorm <- function(x) max(rowSums(abs(x)))
  n <- ncol(M)
  U <- matrix(0, n, n)
  X <- M
  iter <- 0
  converged <- FALSE
  while (iter < maxits && !converged) {
    Y <- X
    T <- Y - U
    e <- eigen(Y, symmetric = TRUE)
    Q <- e$vectors
    d <- e$values
    D <- if (length(d) > 1) 
      diag(d)
    else as.matrix(d)
    p <- (d > eig.tol * d[1])
    QQ <- Q[, p, drop = FALSE]
    X <- QQ %*% D[p, p, drop = FALSE] %*% t(QQ)
    U <- X - T
    X <- (X + t(X))/2
    conv <- inorm(Y - X)/inorm(Y)
    iter <- iter + 1
    converged <- conv <= conv.tol
  }
  X <- (X + t(X))/2
  e <- eigen(X, symmetric = TRUE)
  d <- e$values
  Eps <- posd.tol * abs(d[1])
  if (d[n] < Eps) {
    d[d < Eps] <- Eps
    Q <- e$vectors
    o.diag <- diag(X)
    X <- Q %*% (d * t(Q))
    D <- sqrt(pmax(Eps, o.diag)/diag(X))
    X[] <- D * X * rep(D, each = n)
  }
  (X + t(X))/2
}


H.longPC = function (betas) {
  eta.yx <- as.vector(X %*% betas)
  if (parameterization %in% c("value", "both")) {
    Ys <- as.vector(Xs %*% betas) + Zsb
    Ws.intF.vl.alph <- c(Ws.intF.vl %*% alpha)
    eta.s <- Ws.intF.vl.alph * Ys
  }
  if (parameterization %in% c("slope", "both")) {
    Ys.deriv <- as.vector(Xs.deriv %*% betas[indFixed]) + 
      Zsb.deriv
    Ws.intF.sl.alph <- c(Ws.intF.sl %*% Dalpha)
    eta.s <- if (parameterization == "both") 
      eta.s + Ws.intF.sl.alph * Ys.deriv
    else Ws.intF.sl.alph * Ys.deriv
  }
  exp.eta.tw <- exp(eta.tw)
  H1 <- XtX/sigma^2
  Int <- xi[ind.K] * wkP * exp(eta.s)
  H2 <- H1
  H2 <- matrix(0, ncx, ncx)
  for (i in 1:ncx) {
    for (j in i:ncx) {
      XX <- if (parameterization == "value") {
        Ws.intF.vl.alph^2 * Xs[, i] * Xs[, j]
      }
      else if (parameterization == "slope") {
        if (i %in% indFixed && j %in% indFixed) {
          ii <- match(i, indFixed)
          jj <- match(j, indFixed)
          Ws.intF.sl.alph^2 * Xs.deriv[, ii] * Xs.deriv[, 
                                                        jj]
        }
        else 0
      }
      else {
        if (i %in% indFixed && j %in% indFixed) {
          ii <- match(i, indFixed)
          jj <- match(j, indFixed)
          (Ws.intF.vl.alph * Xs[, i] + Ws.intF.sl.alph * 
            Xs.deriv[, ii]) * (Ws.intF.vl.alph * Xs[, 
                                                    j] + Ws.intF.sl.alph * Xs.deriv[, jj])
        }
        else if (i %in% indFixed && !j %in% indFixed) {
          ii <- match(i, indFixed)
          (Ws.intF.vl.alph * Xs[, i] + Ws.intF.sl.alph * 
            Xs.deriv[, ii]) * (Ws.intF.vl.alph * Xs[, 
                                                    j])
        }
        else if (!i %in% indFixed && j %in% indFixed) {
          jj <- match(j, indFixed)
          (Ws.intF.vl.alph * Xs[, i]) * (Ws.intF.vl.alph * 
                                           Xs[, j] + Ws.intF.sl.alph * Xs.deriv[, jj])
        }
        else {
          Ws.intF.vl.alph^2 * Xs[, i] * Xs[, j]
        }
      }
      ki <- exp.eta.tw * rowsum(Int * XX, id.GK, reorder = FALSE)
      kii <- c((p.byt * ki) %*% wGH)
      H2[i, j] <- sum(kii, na.rm = TRUE)
    }
  }
  H2[lower.tri(H2)] <- t(H2)[lower.tri(H2)]
  H1 + H2
}

gr.longPC = function (betas) {
  eta.yx <- as.vector(X %*% betas)
  if (parameterization %in% c("value", "both")) {
    Ys <- as.vector(Xs %*% betas) + Zsb
    WintF.vl.alph <- c(WintF.vl %*% alpha)
    Ws.intF.vl.alph <- c(Ws.intF.vl %*% alpha)
    eta.s <- Ws.intF.vl.alph * Ys
  }
  if (parameterization %in% c("slope", "both")) {
    Ys.deriv <- as.vector(Xs.deriv %*% betas[indFixed]) + 
      Zsb.deriv
    WintF.sl.alph <- c(WintF.sl %*% Dalpha)
    Ws.intF.sl.alph <- c(Ws.intF.sl %*% Dalpha)
    eta.s <- if (parameterization == "both") 
      eta.s + Ws.intF.sl.alph * Ys.deriv
    else Ws.intF.sl.alph * Ys.deriv
  }
  exp.eta.tw <- exp(eta.tw)
  sc1 <- -crossprod(X, y - eta.yx - Zb)/sigma^2
  Int <- xi[ind.K] * wkP * exp(eta.s)
  sc2 <- numeric(ncx)
  for (i in 1:ncx) {
    ki <- exp.eta.tw * switch(parameterization, value = rowsum(Int * 
                                                                 Ws.intF.vl.alph * Xs[, i], id.GK, reorder = FALSE), 
                              slope = {
                                ii <- match(i, indFixed)
                                if (is.na(ii)) 0 else rowsum(Int * Ws.intF.sl.alph * 
                                                               Xs.deriv[, ii], id.GK, reorder = FALSE)
                              }, both = {
                                ii <- match(i, indFixed)
                                rowsum(Int * (Ws.intF.vl.alph * Xs[, i] + Ws.intF.sl.alph * 
                                                if (is.na(ii)) 0 else Xs.deriv[, ii]), id.GK, 
                                       reorder = FALSE)
                              })
    kii <- c((p.byt * ki) %*% wGH)
    sc2[i] <- switch(parameterization, value = -sum(d * WintF.vl.alph * 
                                                      Xtime[, i] - kii, na.rm = TRUE), slope = {
                                                        ii <- match(i, indFixed)
                                                        if (is.na(ii)) 0 else -sum(d * WintF.sl.alph * Xtime.deriv[, 
                                                                                                                   ii] - kii, na.rm = TRUE)
                                                      }, both = {
                                                        ii <- match(i, indFixed)
                                                        -sum(d * (WintF.vl.alph * Xtime[, i] + WintF.sl.alph * 
                                                                    if (is.na(ii)) 0 else Xtime.deriv[, ii]) - kii, 
                                                             na.rm = TRUE)
                                                      })
  }
  c(sc1 + sc2)
}

gr.survPC = function (thetas){
  thetas <- relist(thetas, skeleton = list.thetas)
  gammas <- thetas$gammas
  alpha <- thetas$alpha
  Dalpha <- thetas$Dalpha
  xi <- exp(thetas$log.xi)
  eta.tw <- if (!is.null(WW)) 
    as.vector(WW %*% gammas)
  else rep(0, n)
  eta.t <- switch(parameterization, value = eta.tw + c(WintF.vl %*% 
                                                         alpha) * Y, slope = eta.tw + c(WintF.sl %*% Dalpha) * 
                    Y.deriv, both = eta.tw + c(WintF.vl %*% alpha) * Y + 
                    c(WintF.sl %*% Dalpha) * Y.deriv)
  exp.eta.s <- exp(switch(parameterization, value = c(Ws.intF.vl %*% 
                                                        alpha) * Ys, slope = c(Ws.intF.sl %*% Dalpha) * Ys.deriv, 
                          both = c(Ws.intF.vl %*% alpha) * Ys + c(Ws.intF.sl %*% 
                                                                    Dalpha) * Ys.deriv))
  exp.eta.tw <- exp(eta.tw)
  Int <- wkP * exp.eta.s
  Int2 <- xi[ind.K] * Int
  scgammas <- if (!is.null(WW)) {
    -colSums(WW * (d - c((p.byt * (exp.eta.tw * rowsum(Int2, 
                                                       id.GK, reorder = FALSE))) %*% wGH)), na.rm = TRUE)
  }
  else NULL
  scalpha <- if (parameterization %in% c("value", "both")) {
    rr <- numeric(ncol(WintF.vl))
    for (k in seq_along(rr)) rr[k] <- -sum((p.byt * (d * 
                                                       WintF.vl[, k] * Y - exp.eta.tw * rowsum(Int2 * Ws.intF.vl[, 
                                                                                                                 k] * Ys, id.GK, reorder = FALSE))) %*% wGH, na.rm = TRUE)
    rr
  }
  else NULL
  scalpha.D <- if (parameterization %in% c("slope", "both")) {
    rr <- numeric(ncol(WintF.sl))
    for (k in seq_along(rr)) rr[k] <- -sum((p.byt * (d * 
                                                       WintF.sl[, k] * Y.deriv - exp.eta.tw * rowsum(Int2 * 
                                                                                                       Ws.intF.sl[, k] * Ys.deriv, id.GK, reorder = FALSE))) %*% 
                                             wGH, na.rm = TRUE)
    rr
  }
  else NULL
  scxi <- numeric(Q)
  for (i in 1:Q) {
    i1 <- ind.D == i
    i2 <- ind.K == i
    i3 <- ind.D >= i
    ki <- c((p.byt[i3, ] * (exp.eta.tw[i3] * rowsum(Int[i2, 
                                                        ], id.GK[i2], reorder = FALSE))) %*% wGH)
    kk <- numeric(n)
    kk[i3] <- ki
    scxi[i] <- -xi[i] * sum((d * i1)/xi[i] - kk)
  }
  c(scgammas, scalpha, scalpha.D, scxi)
}

opt.survPC = function (thetas) {
  thetas <- relist(thetas, skeleton = list.thetas)
  gammas <- thetas$gammas
  alpha <- thetas$alpha
  Dalpha <- thetas$Dalpha
  xi <- exp(thetas$log.xi)
  eta.tw <- if (!is.null(WW)) 
    as.vector(WW %*% gammas)
  else 0
  eta.t <- switch(parameterization, value = eta.tw + c(WintF.vl %*% 
                                                         alpha) * Y, slope = eta.tw + c(WintF.sl %*% Dalpha) * 
                    Y.deriv, both = eta.tw + c(WintF.vl %*% alpha) * Y + 
                    c(WintF.sl %*% Dalpha) * Y.deriv)
  eta.s <- switch(parameterization, value = c(Ws.intF.vl %*% 
                                                alpha) * Ys, slope = c(Ws.intF.sl %*% Dalpha) * Ys.deriv, 
                  both = c(Ws.intF.vl %*% alpha) * Ys + c(Ws.intF.sl %*% 
                                                            Dalpha) * Ys.deriv)
  log.hazard <- log(xi[ind.D]) + eta.t
  log.survival <- -exp(eta.tw) * rowsum(xi[ind.K] * wkP * exp(eta.s), 
                                        id.GK, reorder = FALSE)
  dimnames(log.survival) <- NULL
  log.p.tb <- d * log.hazard + log.survival
  p.bytn <- p.byt * log.p.tb
  -sum(p.bytn %*% wGH, na.rm = TRUE)
}

chol.transf = function (x){
  if (any(is.na(x) | !is.finite(x))) 
    stop("NA or infinite values in 'x'.\n")
  if (is.matrix(x)) {
    k <- nrow(x)
    U <- chol(x)
    U[cbind(1:k, 1:k)] <- log(U[cbind(1:k, 1:k)])
    U[upper.tri(U, TRUE)]
  }
  else {
    nx <- length(x)
    k <- round((-1 + sqrt(1 + 8 * nx))/2)
    mat <- matrix(0, k, k)
    mat[upper.tri(mat, TRUE)] <- x
    mat[cbind(1:k, 1:k)] <- exp(mat[cbind(1:k, 1:k)])
    res <- crossprod(mat)
    attr(res, "L") <- t(mat)[lower.tri(mat, TRUE)]
    res
  }
}

LogLik.piecewiseGH = function (thetas){
  thetas <- relist(thetas, skeleton = list.thetas)
  betas <- thetas$betas
  sigma <- exp(thetas$log.sigma)
  gammas <- thetas$gammas
  alpha <- thetas$alpha
  Dalpha <- thetas$Dalpha
  xi <- exp(thetas$log.xi)
  D <- thetas$D
  D <- if (diag.D) 
    exp(D)
  else chol.transf(D)
  eta.yx <- as.vector(X %*% betas)
  eta.tw <- if (!is.null(WW)) 
    as.vector(WW %*% gammas)
  else 0
  if (parameterization %in% c("value", "both")) {
    Y <- as.vector(Xtime %*% betas) + Ztime.b
    Ys <- as.vector(Xs %*% betas) + Zsb
    eta.t <- eta.tw + c(WintF.vl %*% alpha) * Y
    eta.s <- c(Ws.intF.vl %*% alpha) * Ys
  }
  if (parameterization %in% c("slope", "both")) {
    Y.deriv <- as.vector(Xtime.deriv %*% betas[indFixed]) + 
      Ztime.b.deriv
    Ys.deriv <- as.vector(Xs.deriv %*% betas[indFixed]) + 
      Zsb.deriv
    eta.t <- if (parameterization == "both") 
      eta.t + c(WintF.sl %*% Dalpha) * Y.deriv
    else eta.tw + c(WintF.sl %*% Dalpha) * Y.deriv
    eta.s <- if (parameterization == "both") 
      eta.s + c(Ws.intF.sl %*% Dalpha) * Ys.deriv
    else c(Ws.intF.sl %*% Dalpha) * Ys.deriv
  }
  mu.y <- eta.yx + Ztb
  logNorm <- dnorm(y, mu.y, sigma, TRUE)
  log.p.yb <- rowsum(logNorm, id)
  log.hazard <- log(xi[ind.D]) + eta.t
  log.survival <- -exp(eta.tw) * rowsum(xi[ind.K] * wkP * exp(eta.s), 
                                        id.GK, reorder = FALSE)
  dimnames(log.survival) <- NULL
  log.p.tb <- d * log.hazard + log.survival
  log.p.b <- if (control$typeGH == "simple") {
    rr <- dmvnorm(b, rep(0, ncz), D, TRUE)
    rep(rr, each = n)
  }
  else {
    matrix(dmvnorm(do.call(rbind, lis.b), rep(0, ncz), D, 
                   TRUE), n, k, byrow = TRUE)
  }
  p.ytb <- exp(log.p.yb + log.p.tb + log.p.b)
  if (control$typeGH != "simple") 
    p.ytb <- p.ytb * VCdets
  dimnames(p.ytb) <- NULL
  p.yt <- c(p.ytb %*% wGH)
  log.p.yt <- log(p.yt)
  -sum(log.p.yt[is.finite(log.p.yt)], na.rm = TRUE)
}

Score.piecewiseGH = function (thetas) 
{
  thetas <- relist(thetas, skeleton = list.thetas)
  betas <- thetas$betas
  sigma <- exp(thetas$log.sigma)
  gammas <- thetas$gammas
  alpha <- thetas$alpha
  Dalpha <- thetas$Dalpha
  xi <- exp(thetas$log.xi)
  D <- thetas$D
  D <- if (diag.D) 
    exp(D)
  else chol.transf(D)
  eta.yx <- as.vector(X %*% betas)
  eta.tw <- if (!is.null(WW)) 
    as.vector(WW %*% gammas)
  else rep(0, n)
  if (parameterization %in% c("value", "both")) {
    Y <- as.vector(Xtime %*% betas) + Ztime.b
    Ys <- as.vector(Xs %*% betas) + Zsb
    WintF.vl.alph <- c(WintF.vl %*% alpha)
    Ws.intF.vl.alph <- c(Ws.intF.vl %*% alpha)
    eta.t <- eta.tw + WintF.vl.alph * Y
    eta.s <- Ws.intF.vl.alph * Ys
  }
  if (parameterization %in% c("slope", "both")) {
    Y.deriv <- as.vector(Xtime.deriv %*% betas[indFixed]) + 
      Ztime.b.deriv
    Ys.deriv <- as.vector(Xs.deriv %*% betas[indFixed]) + 
      Zsb.deriv
    WintF.sl.alph <- c(WintF.sl %*% Dalpha)
    Ws.intF.sl.alph <- c(Ws.intF.sl %*% Dalpha)
    eta.t <- if (parameterization == "both") 
      eta.t + WintF.sl.alph * Y.deriv
    else eta.tw + WintF.sl.alph * Y.deriv
    eta.s <- if (parameterization == "both") 
      eta.s + Ws.intF.sl.alph * Ys.deriv
    else Ws.intF.sl.alph * Ys.deriv
  }
  exp.eta.tw <- exp(eta.tw)
  exp.eta.s <- exp(eta.s)
  mu.y <- eta.yx + Ztb
  logNorm <- dnorm(y, mu.y, sigma, TRUE)
  log.p.yb <- rowsum(logNorm, id)
  log.hazard <- log(xi[ind.D]) + eta.t
  Int <- xi[ind.K] * wkP * exp.eta.s
  log.survival <- -exp(eta.tw) * rowsum(Int, id.GK, reorder = FALSE)
  dimnames(log.survival) <- NULL
  log.p.tb <- d * log.hazard + log.survival
  log.p.b <- if (control$typeGH == "simple") {
    rr <- dmvnorm(b, rep(0, ncz), D, TRUE)
    rep(rr, each = n)
  }
  else {
    matrix(dmvnorm(do.call(rbind, lis.b), rep(0, ncz), D, 
                   TRUE), n, k, byrow = TRUE)
  }
  p.ytb <- exp(log.p.yb + log.p.tb + log.p.b)
  if (control$typeGH != "simple") 
    p.ytb <- p.ytb * VCdets
  dimnames(p.ytb) <- NULL
  p.yt <- c(p.ytb %*% wGH)
  p.byt <- p.ytb/p.yt
  post.b <- if (control$typeGH == "simple") {
    p.byt %*% (b * wGH)
  }
  else {
    sapply(seq_len(ncz), function(i) (p.byt * t(sapply(lis.b, 
                                                       "[", seq_len(k), i))) %*% wGH)
  }
  post.vb <- if (control$typeGH == "simple") {
    if (ncz == 1) {
      c(p.byt %*% (b2 * wGH)) - c(post.b * post.b)
    }
    else {
      (p.byt %*% (b2 * wGH)) - t(apply(post.b, 1, function(x) x %o% 
                                         x))
    }
  }
  else {
    dd <- sapply(seq_len(ncz^2), function(i) (p.byt * t(sapply(lis.b2, 
                                                               "[", seq_len(k), i))) %*% wGH)
    bb <- apply(post.b, 1, function(x) x %o% x)
    dd - if (ncz == 1) 
      c(bb)
    else t(bb)
  }
  Zb <- if (ncz == 1) 
    post.b[id]
  else rowSums(Z * post.b[id, ], na.rm = TRUE)
  mu <- y - eta.yx
  tr.tZZvarb <- sum(ZtZ * post.vb, na.rm = TRUE)
  sc1 <- -crossprod(X, y - eta.yx - Zb)/sigma^2
  sc2 <- numeric(ncx)
  for (i in 1:ncx) {
    ki <- exp.eta.tw * switch(parameterization, value = rowsum(Int * 
                                                                 Ws.intF.vl.alph * Xs[, i], id.GK, reorder = FALSE), 
                              slope = {
                                ii <- match(i, indFixed)
                                if (is.na(ii)) 0 else rowsum(Int * Ws.intF.sl.alph * 
                                                               Xs.deriv[, ii], id.GK, reorder = FALSE)
                              }, both = {
                                ii <- match(i, indFixed)
                                rowsum(Int * (Ws.intF.vl.alph * Xs[, i] + Ws.intF.sl.alph * 
                                                if (is.na(ii)) 0 else Xs.deriv[, ii]), id.GK, 
                                       reorder = FALSE)
                              })
    kii <- c((p.byt * ki) %*% wGH)
    sc2[i] <- switch(parameterization, value = -sum(d * WintF.vl.alph * 
                                                      Xtime[, i] - kii, na.rm = TRUE), slope = {
                                                        ii <- match(i, indFixed)
                                                        if (is.na(ii)) 0 else -sum(d * WintF.sl.alph * Xtime.deriv[, 
                                                                                                                   ii] - kii, na.rm = TRUE)
                                                      }, both = {
                                                        ii <- match(i, indFixed)
                                                        -sum(d * (WintF.vl.alph * Xtime[, i] + WintF.sl.alph * 
                                                                    if (is.na(ii)) 0 else Xtime.deriv[, ii]) - kii, 
                                                             na.rm = TRUE)
                                                      })
  }
  score.y <- c(sc1 + sc2, -sigma * (-N/sigma + drop(crossprod(mu, 
                                                              mu - 2 * Zb) + crossprod(Zb) + tr.tZZvarb)/sigma^3))
  Int2 <- wkP * exp.eta.s
  scgammas <- if (!is.null(WW)) {
    -colSums(WW * (d - c((p.byt * (exp.eta.tw * rowsum(Int, 
                                                       id.GK, reorder = FALSE))) %*% wGH)), na.rm = TRUE)
  }
  else NULL
  scalpha <- if (parameterization %in% c("value", "both")) {
    rr <- numeric(ncol(WintF.vl))
    for (l in seq_along(rr)) rr[l] <- -sum((p.byt * (d * 
                                                       WintF.vl[, l] * Y - exp.eta.tw * rowsum(Int * Ws.intF.vl[, 
                                                                                                                l] * Ys, id.GK, reorder = FALSE))) %*% wGH, na.rm = TRUE)
    rr
  }
  else NULL
  scalpha.D <- if (parameterization %in% c("slope", "both")) {
    rr <- numeric(ncol(WintF.sl))
    for (l in seq_along(rr)) rr[l] <- -sum((p.byt * (d * 
                                                       WintF.sl[, l] * Y.deriv - exp.eta.tw * rowsum(Int * 
                                                                                                       Ws.intF.sl[, l] * Ys.deriv, id.GK, reorder = FALSE))) %*% 
                                             wGH, na.rm = TRUE)
    rr
  }
  else NULL
  scxi <- numeric(Q)
  for (i in 1:Q) {
    i1 <- ind.D == i
    i2 <- ind.K == i
    i3 <- ind.D >= i
    ki <- c((p.byt[i3, ] * (exp.eta.tw[i3] * rowsum(Int2[i2, 
                                                         ], id.GK[i2], reorder = FALSE))) %*% wGH)
    kk <- numeric(n)
    kk[i3] <- ki
    scxi[i] <- -xi[i] * sum((d * i1)/xi[i] - kk)
  }
  score.t <- c(scgammas, scalpha, scalpha.D, scxi)
  score.b <- if (diag.D) {
    svD <- 1/D
    svD2 <- svD^2
    cS.postVB <- colSums(as.matrix(post.vb), na.rm = TRUE)
    dim(cS.postVB) <- c(ncz, ncz)
    D * 0.5 * (n * svD - diag(cS.postVB) * svD2 - colSums(as.matrix(post.b^2), 
                                                          na.rm = TRUE) * svD2)
  }
  else {
    svD <- solve(D)
    dD <- deriv.D(D)
    ndD <- length(dD)
    D1 <- sapply(dD, function(x) sum(svD * x))
    D2 <- t(sapply(dD, function(x) c(svD %*% x %*% svD)))
    cS.postVB <- colSums(as.matrix(post.vb), na.rm = TRUE)
    out <- numeric(ndD)
    for (i in seq_along(dD)) {
      D.mat <- D2[i, ]
      dim(D.mat) <- c(ncz, ncz)
      out[i] <- sum(D2[i, ] * cS.postVB, na.rm = TRUE) + 
        sum((post.b %*% D.mat) * post.b, na.rm = TRUE)
    }
    J <- jacobian2(attr(D, "L"), ncz)
    drop(0.5 * (n * D1 - out) %*% J)
  }
  c(score.y, score.t, score.b)
}


fd.vec = function (x, f, ..., eps = 1e-05) 
{
  n <- length(x)
  res <- matrix(0, n, n)
  ex <- pmax(abs(x), 1)
  f0 <- f(x, ...)
  for (i in 1:n) {
    x1 <- x
    x1[i] <- x[i] + eps * ex[i]
    diff.f <- c(f(x1, ...) - f0)
    diff.x <- x1[i] - x[i]
    res[, i] <- diff.f/diff.x
  }
  0.5 * (res + t(res))
}


dropAttr = function (mat) {
  d <- dim(mat)
  mat <- as.vector(mat)
  dim(mat) <- d
  mat
}


summary.myFJM = function (object, ...) 
{
  if (!inherits(object, "FJM")) 
    stop("Use only with 'FJM' objects.\n")
  
  VarCov <- vcov(object)
  betas <- object$coefficients$betas
  indY <- seq(1, length(betas))
  sds <- sqrt(diag(VarCov[indY, indY]))
  coefsY <- cbind(Value = betas, Std.Err = sds, `z-value` = betas/sds, 
                  `p-value` = 2 * pnorm(abs(betas/sds), lower.tail = FALSE))
  if (object$method == "weibull-PH-GH") {
    gammas <- c(object$coefficients$gammas, Assoct = as.vector(object$coefficients$alpha), 
                Assoct.s = as.vector(object$coefficients$Dalpha), 
                `log(shape)` = log(as.vector(object$coefficients$sigma.t)))
    indT <- grep("T.", colnames(VarCov), fixed = TRUE)
  }else if (object$method == "piecewise-PH-GH") {
    gammas <- c(object$coefficients$gammas, Assoct = as.vector(object$coefficients$alpha), 
                Assoct.s = as.vector(object$coefficients$Dalpha), 
                log(as.vector(object$coefficients$xi)))
    ss <- seq(length(gammas) - object$x$Q + 1, length(gammas))
    names(gammas)[ss] <- paste("log(xi.", seq_len(object$x$Q), 
                               ")", sep = "")
    indT <- grep("T.", colnames(VarCov), fixed = TRUE)
  }
  jj <- grep("Assoct[!^\\.s]", names(gammas))
  ii <- setdiff(grep("Assoct", names(gammas)), jj)
  if (length(ii) > 1) {
    nn <- names(object$coefficients$alpha)
    names(gammas)[ii] <- if (length(nn) == 1) 
      "Assoct"
    else {
      if (nn[1] == "") 
        c("Assoct", paste("Assoct", nn[-1], sep = ":"))
      else paste("Assoct", nn, sep = ":")
    }
  }
  if (length(jj) > 1) {
    nn <- names(object$coefficients$Dalpha)
    names(gammas)[jj] <- if (length(nn) == 1) 
      "Assoct.s"
    else {
      if (nn[1] == "") 
        c("Assoct.s", paste("Assoct.s", nn[-1], sep = ":"))
      else paste("Assoct.s", nn, sep = ":")
    }
  }
  if ((lag <- object$y$lag) > 0) {
    kk <- grep("Assoct", names(gammas), fixed = TRUE)
    names(gammas)[kk] <- paste(names(gammas)[kk], "(lag=", 
                               lag, ")", sep = "")
  }
  sds <- if (length(indT) > 1) 
    sqrt(diag(VarCov[indT, indT]))
  else sqrt(VarCov[indT, indT])
  if (!is.null(object$scaleWB)) 
    sds <- c(sds, NA)
  coefsT <- cbind(Value = gammas, Std.Err = sds, `z-value` = gammas/sds, 
                  `p-value` = 2 * pnorm(abs(gammas/sds), lower.tail = FALSE))
  out <- list(`CoefTable-Long` = coefsY, `CoefTable-Event` = coefsT, 
              D = object$coefficients$D, sigma = object$coefficients$sigma, 
              logLik = as.vector(logLik(object)), AIC = AIC(object), 
              BIC = AIC(object, k = log(object$n)))
  out$N <- object$N
  out$n <- object$n
  out$d <- object$d
  out$id <- object$id
  out$method <- object$method
  out$control <- object$control
  out$knots <- unique(object$knots)
  out$conv <- object$conv
  out$parameterization <- object$parameterization
  out$nstrata <- length(unique(object$y$strata))
  out$CompRisk <- object$CompRisk
  out$call <- object$call
  class(out) <- "summary.myFJM"
  out
}

vcov = function (object, ...) 
{
  out <- try(solve(object$Hessian), silent = TRUE)
  vmat <- if (!inherits(out, "try-error")) 
    structure(out, dimnames = dimnames(object$Hessian))
  else structure(ginv(object$Hessian), dimnames = dimnames(object$Hessian))
  (vmat + t(vmat))/2
}

logLik = function (object, ...) 
{
  if (!inherits(object, "FJM")) 
    stop("Use only with 'FJM' objects.\n")
  out <- object$logLik
  attr(out, "df") <- nrow(object$Hessian)
  attr(out, "nobs") <- object$n
  class(out) <- "logLik"
  out
}

 
log.posterior.b = function (b, y, Mats, method, ii) 
{
  id.i <- id %in% ii
  idT.i <- idT %in% ii
  X.i <- X[id.i, , drop = FALSE]
  Z.i <- Z[id.i, , drop = FALSE]
  mu.y <- as.vector(X.i %*% betas.new) + rowSums(Z.i * rep(b, 
                                                           each = nrow(Z.i)))
  logNorm <- dnorm(y[id.i], mu.y, sigma.new, TRUE)
  log.p.yb <- sum(logNorm)
  log.p.b <- dmvnorm(b, rep(0, ncol(Z)), D.new, TRUE)
  st <- Mats[[ii]]$st
  wk <- Mats[[ii]]$wk
  P <- Mats[[ii]]$P
  Xs <- Mats[[ii]]$Xs
  Zs <- Mats[[ii]]$Zs
  Xs.deriv <- Mats[[ii]]$Xs.deriv
  Zs.deriv <- Mats[[ii]]$Zs.deriv
  Ws.intF.vl <- Mats[[ii]]$Ws.intF.vl
  Ws.intF.sl <- Mats[[ii]]$Ws.intF.sl
  ind <- Mats[[ii]]$ind
  if (parameterization %in% c("value", "both")) 
    Ys <- as.vector(Xs %*% betas.new + rowSums(Zs * rep(b, 
                                                        each = nrow(Zs))))
  if (parameterization %in% c("slope", "both")) 
    Ys.deriv <- as.vector(Xs.deriv %*% betas.new[indFixed]) + 
    rowSums(Zs.deriv * rep(b[indRandom], each = nrow(Zs)))
  tt <- switch(parameterization, value = c(Ws.intF.vl %*% alpha.new) * 
                 Ys, slope = c(Ws.intF.sl %*% Dalpha.new) * Ys.deriv, 
               both = c(Ws.intF.vl %*% alpha.new) * Ys + c(Ws.intF.sl %*% 
                                                             Dalpha.new) * Ys.deriv)
  eta.tw <- if (!is.null(W)) {
    if (!LongFormat) 
      as.vector(W[ii, , drop = FALSE] %*% gammas.new)
    else as.vector(W[idT.i, , drop = FALSE] %*% gammas.new)
  }
  else 0
  log.survival <- if (method == "weibull-PH-GH") {
    Vi <- exp(log(sigma.t.new) + (sigma.t.new - 1) * log(st) + 
                tt)
    -exp(eta.tw) * P * sum(wk * Vi)
  }
  else if (method == "weibull-AFT-GH") {
    Vi <- exp(eta.tw) * P * sum(wk * exp(tt))
    -Vi^sigma.t.new
  }
  else if (method == "spline-PH-GH") {
    W2s <- if (length(kn <- object$control$knots) == 1) {
      splineDesign(unlist(kn, use.names = FALSE), st, ord = object$control$ord, 
                   outer.ok = TRUE)
    }
    else {
      strt.i <- strt[ii]
      w2s <- lapply(kn, function(kn) splineDesign(kn, st, 
                                                  ord = object$control$ord, outer.ok = TRUE))
      ll <- match(strt.i, names(w2s))
      w2s[-ll] <- lapply(w2s[-ll], function(m) {
        m[, ] <- 0
        m
      })
      do.call(cbind, w2s)
    }
    Vi <- exp(c(W2s %*% gammas.bs.new) + tt)
    idT <- rep(seq_along(P), each = object$control$GKk)
    -sum(exp(eta.tw) * P * tapply(wk * Vi, idT, sum))
  }
  else if (method == "piecewise-PH-GH") {
    P <- P[!is.na(P)]
    ind.K <- rep(seq_len(ind), each = 7)
    wk <- rep(wk, ind)
    wkP <- wk * rep(P, each = 7)
    eta.tw <- if (!is.null(W)) 
      as.vector(W[i, , drop = FALSE] %*% gammas.new)
    else 0
    -exp(eta.tw) * sum(xi.new[ind.K] * wkP * exp(tt))
  }
  log.p.yb + log.survival + log.p.b
}


S.b  = function (t, b, ii, Mats) 
{
  if (t == 0) 
    return(1)
  idT.i <- idT %in% ii
  st <- Mats$st
  wk <- Mats$wk
  P <- Mats$P
  Xs <- Mats$Xs
  Zs <- Mats$Zs
  Xs.deriv <- Mats$Xs.deriv
  Zs.deriv <- Mats$Zs.deriv
  Ws.intF.vl <- Mats$Ws.intF.vl
  Ws.intF.sl <- Mats$Ws.intF.sl
  ind <- Mats$ind
  if (parameterization %in% c("value", "both")) 
    Ys <- as.vector(Xs %*% betas.new + rowSums(Zs * rep(b, 
                                                        each = nrow(Zs))))
  if (parameterization %in% c("slope", "both")) 
    Ys.deriv <- as.vector(Xs.deriv %*% betas.new[indFixed]) + 
    rowSums(Zs.deriv * rep(b[indRandom], each = nrow(Zs)))
  tt <- switch(parameterization, value = c(Ws.intF.vl %*% alpha.new) * 
                 Ys, slope = c(Ws.intF.sl %*% Dalpha.new) * Ys.deriv, 
               both = c(Ws.intF.vl %*% alpha.new) * Ys + c(Ws.intF.sl %*% 
                                                             Dalpha.new) * Ys.deriv)
  eta.tw <- if (!is.null(W)) {
    if (!LongFormat) 
      as.vector(W[ii, , drop = FALSE] %*% gammas.new)
    else as.vector(W[idT.i %in% ii, , drop = FALSE] %*% gammas.new)
  }
  else 0
  log.survival <- if (method == "weibull-PH-GH") {
    Vi <- exp(log(sigma.t.new) + (sigma.t.new - 1) * log(st) + 
                tt)
    -exp(eta.tw) * P * sum(wk * Vi)
  }
  else if (method == "weibull-AFT-GH") {
    Vi <- exp(eta.tw) * P * sum(wk * exp(tt))
    -Vi^sigma.t.new
  }
  else if (method == "spline-PH-GH") {
    W2s <- if (length(kn <- object$control$knots) == 1) {
      splineDesign(unlist(kn, use.names = FALSE), st, ord = object$control$ord, 
                   outer.ok = TRUE)
    }
    else {
      strt.i <- strt[ii]
      w2s <- lapply(kn, function(kn) splineDesign(kn, st, 
                                                  ord = object$control$ord, outer.ok = TRUE))
      ll <- match(strt.i, names(w2s))
      w2s[-ll] <- lapply(w2s[-ll], function(m) {
        m[, ] <- 0
        m
      })
      do.call(cbind, w2s)
    }
    Vi <- exp(c(W2s %*% gammas.bs.new) + tt)
    idT <- rep(seq_along(P), each = object$control$GKk)
    -sum(exp(eta.tw) * P * tapply(wk * Vi, idT, sum))
  }
  else if (method == "piecewise-PH-GH") {
    P <- P[!is.na(P)]
    ind.K <- rep(seq_len(ind), each = 7)
    wk <- rep(wk, ind)
    wkP <- wk * rep(P, each = 7)
    eta.tw <- if (!is.null(W)) 
      as.vector(W[i, , drop = FALSE] %*% gammas.new)
    else 0
    -exp(eta.tw) * sum(xi.new[ind.K] * wkP * exp(tt))
  }
  exp(log.survival)
}

opt.survWB = function (thetas) {
  thetas <- relist(thetas, skeleton = list.thetas)
  gammas <- thetas$gammas
  alpha <- thetas$alpha
  Dalpha <- thetas$Dalpha
  sigma.t <- if (is.null(scaleWB)) 
    exp(thetas$log.sigma.t)
  else scaleWB
  eta.tw <- as.vector(WW %*% gammas)
  eta.t <- switch(parameterization, value = eta.tw + c(WintF.vl %*% alpha) * Y, slope = eta.tw + c(WintF.sl %*% Dalpha) * 
                    Y.deriv, both = eta.tw + c(WintF.vl %*% alpha) * Y + 
                    c(WintF.sl %*% Dalpha) * Y.deriv)
  eta.s <- switch(parameterization, value = c(Ws.intF.vl %*%  alpha) * Ys, slope = c(Ws.intF.sl %*% Dalpha) * Ys.deriv, 
                  both = c(Ws.intF.vl %*% alpha) * Ys + c(Ws.intF.sl %*%  Dalpha) * Ys.deriv)
  log.hazard <- log(sigma.t) + (sigma.t - 1) * logT + eta.t
  log.survival <- -exp(eta.tw) * P * rowsum(wk * exp(log(sigma.t) +  (sigma.t - 1) * log.st + eta.s), id.GK, reorder = FALSE)
  dimnames(log.survival) <- NULL
  log.p.tb <- d * log.hazard + log.survival
  p.bytn <- p.byt * log.p.tb
  -sum(p.bytn %*% wGH, na.rm = TRUE)
}

gr.survWB = function (thetas) {
  thetas <- relist(thetas, skeleton = list.thetas)
  gammas <- thetas$gammas
  alpha <- thetas$alpha
  Dalpha <- thetas$Dalpha
  sigma.t <- if (is.null(scaleWB)) 
    exp(thetas$log.sigma.t)
  else scaleWB
  eta.tw <- as.vector(WW %*% gammas)
  eta.t <- switch(parameterization, value = eta.tw + c(WintF.vl %*%  alpha) * Y, slope = eta.tw + c(WintF.sl %*% Dalpha) * 
                    Y.deriv, both = eta.tw + c(WintF.vl %*% alpha) * Y +  c(WintF.sl %*% Dalpha) * Y.deriv)
  eta.s <- switch(parameterization, value = c(Ws.intF.vl %*%  alpha) * Ys, slope = c(Ws.intF.sl %*% Dalpha) * Ys.deriv, 
                  both = c(Ws.intF.vl %*% alpha) * Ys + c(Ws.intF.sl %*%  Dalpha) * Ys.deriv)
  exp.eta.tw.P <- exp(eta.tw) * P
  Int <- wk * exp(log(sigma.t) + (sigma.t - 1) * log.st + eta.s)
  ki <- exp.eta.tw.P * rowsum(Int, id.GK, reorder = FALSE)
  kii <- c((p.byt * ki) %*% wGH)
  scgammas <- -colSums(WW * (d - kii), na.rm = TRUE)
  scalpha <- if (parameterization %in% c("value", "both")) {
    rr <- numeric(ncol(WintF.vl))
    for (k in seq_along(rr)) rr[k] <- -sum((p.byt * (d * WintF.vl[, k] * Y - exp.eta.tw.P * 
                                                       rowsum(Int * Ws.intF.vl[,k] * Ys, id.GK, reorder = FALSE))) %*% wGH, na.rm = TRUE)
    rr
  }
  else NULL
  scalpha.D <- if (parameterization %in% c("slope", "both")) {
    rr <- numeric(ncol(WintF.sl))
    for (k in seq_along(rr)) rr[k] <- -sum((p.byt * (d *  WintF.sl[, k] * Y.deriv - exp.eta.tw.P * 
                                                       rowsum(Int *Ws.intF.sl[, k] * Ys.deriv, id.GK, reorder = FALSE))) %*% wGH, na.rm = TRUE)
    rr
  }
  else NULL
  scsigmat <- if (is.null(scaleWB)) {
    Int2 <- st^(sigma.t - 1) * (1 + sigma.t * log.st) * exp(eta.s)
    -sigma.t * sum((p.byt * (d * (1/sigma.t + logT) - exp.eta.tw.P *  rowsum(wk * Int2, id.GK, reorder = FALSE))) %*% wGH,  na.rm = TRUE)
  }
  else NULL
  c(scgammas, scalpha, scalpha.D, scsigmat)
}

opt.longWB = function (betas) {
  eta.yx <- as.vector(X %*% betas)
  if (parameterization %in% c("value", "both")) {
    Y <- as.vector(Xtime %*% betas) + Ztime.b
    Ys <- as.vector(Xs %*% betas) + Zsb
    eta.t <- eta.tw + c(WintF.vl %*% alpha) * Y
    eta.s <- c(Ws.intF.vl %*% alpha) * Ys
  }
  if (parameterization %in% c("slope", "both")) {
    Y.deriv <- as.vector(Xtime.deriv %*% betas[indFixed]) + 
      Ztime.b.deriv
    Ys.deriv <- as.vector(Xs.deriv %*% betas[indFixed]) + 
      Zsb.deriv
    eta.t <- if (parameterization == "both") 
      eta.t + c(WintF.sl %*% Dalpha) * Y.deriv
    else eta.tw + c(WintF.sl %*% Dalpha) * Y.deriv
    eta.s <- if (parameterization == "both") 
      eta.s + c(Ws.intF.sl %*% Dalpha) * Ys.deriv
    else c(Ws.intF.sl %*% Dalpha) * Ys.deriv
  }
  mu.y <- eta.yx + Ztb
  logNorm <- dnorm(y, mu.y, sigma, TRUE)
  log.p.yb <- rowsum(logNorm, id)
  log.hazard <- log(sigma.t) + (sigma.t - 1) * logT + eta.t
  log.survival <- -exp(eta.tw) * P * rowsum(wk * exp(log(sigma.t) + 
                                                       (sigma.t - 1) * log.st + eta.s), id.GK, reorder = FALSE)
  log.p.tb <- d * log.hazard + log.survival
  p.bytn <- p.byt * (log.p.yb + log.p.tb)
  -sum(p.bytn %*% wGH, na.rm = TRUE)
}

gr.longWB = function (betas){
  eta.yx <- as.vector(X %*% betas)
  if (parameterization %in% c("value", "both")) {
    Ys <- as.vector(Xs %*% betas) + Zsb
    WintF.vl.alph <- c(WintF.vl %*% alpha)
    Ws.intF.vl.alph <- c(Ws.intF.vl %*% alpha)
    eta.s <- Ws.intF.vl.alph * Ys
  }
  if (parameterization %in% c("slope", "both")) {
    Ys.deriv <- as.vector(Xs.deriv %*% betas[indFixed]) + 
      Zsb.deriv
    WintF.sl.alph <- c(WintF.sl %*% Dalpha)
    Ws.intF.sl.alph <- c(Ws.intF.sl %*% Dalpha)
    eta.s <- if (parameterization == "both") 
      eta.s + Ws.intF.sl.alph * Ys.deriv
    else Ws.intF.sl.alph * Ys.deriv
  }
  exp.eta.tw.P <- exp(eta.tw) * P
  sc1 <- -crossprod(X, y - eta.yx - Zb)/sigma^2
  Int <- wk * exp(log(sigma.t) + (sigma.t - 1) * log.st + eta.s)
  sc2 <- numeric(ncx)
  for (i in 1:ncx) {
    ki <- exp.eta.tw.P * switch(parameterization, value = rowsum(Int * 
                                                                   Ws.intF.vl.alph * Xs[, i], id.GK, reorder = FALSE), 
                                slope = {
                                  ii <- match(i, indFixed)
                                  if (is.na(ii)) 0 else rowsum(Int * Ws.intF.sl.alph * 
                                                                 Xs.deriv[, ii], id.GK, reorder = FALSE)
                                }, both = {
                                  ii <- match(i, indFixed)
                                  rowsum(Int * (Ws.intF.vl.alph * Xs[, i] + Ws.intF.sl.alph * 
                                                  if (is.na(ii)) 0 else Xs.deriv[, ii]), id.GK, 
                                         reorder = FALSE)
                                })
    kii <- c((p.byt * ki) %*% wGH)
    sc2[i] <- switch(parameterization, value = -sum(d * WintF.vl.alph * 
                                                      Xtime[, i] - kii, na.rm = TRUE), slope = {
                                                        ii <- match(i, indFixed)
                                                        if (is.na(ii)) 0 else -sum(d * WintF.sl.alph * Xtime.deriv[, 
                                                                                                                   ii] - kii, na.rm = TRUE)
                                                      }, both = {
                                                        ii <- match(i, indFixed)
                                                        -sum(d * (WintF.vl.alph * Xtime[, i] + WintF.sl.alph * 
                                                                    if (is.na(ii)) 0 else Xtime.deriv[, ii]) - kii, 
                                                             na.rm = TRUE)
                                                      })
  }
  c(sc1 + sc2)
}

H.longWB = function (betas) {
  eta.yx <- as.vector(X %*% betas)
  if (parameterization %in% c("value", "both")) {
    Ys <- as.vector(Xs %*% betas) + Zsb
    Ws.intF.vl.alph <- c(Ws.intF.vl %*% alpha)
    eta.s <- Ws.intF.vl.alph * Ys
  }
  if (parameterization %in% c("slope", "both")) {
    Ys.deriv <- as.vector(Xs.deriv %*% betas[indFixed]) + 
      Zsb.deriv
    Ws.intF.sl.alph <- c(Ws.intF.sl %*% Dalpha)
    eta.s <- if (parameterization == "both") 
      eta.s + Ws.intF.sl.alph * Ys.deriv
    else Ws.intF.sl.alph * Ys.deriv
  }
  exp.eta.tw.P <- exp(eta.tw) * P
  H1 <- XtX/sigma^2
  Int <- wk * exp(log(sigma.t) + (sigma.t - 1) * log.st + eta.s)
  H2 <- matrix(0, ncx, ncx)
  for (i in 1:ncx) {
    for (j in i:ncx) {
      XX <- if (parameterization == "value") {
        Ws.intF.vl.alph^2 * Xs[, i] * Xs[, j]
      }
      else if (parameterization == "slope") {
        if (i %in% indFixed && j %in% indFixed) {
          ii <- match(i, indFixed)
          jj <- match(j, indFixed)
          Ws.intF.sl.alph^2 * Xs.deriv[, ii] * Xs.deriv[, 
                                                        jj]
        }
        else 0
      }
      else {
        if (i %in% indFixed && j %in% indFixed) {
          ii <- match(i, indFixed)
          jj <- match(j, indFixed)
          (Ws.intF.vl.alph * Xs[, i] + Ws.intF.sl.alph * 
            Xs.deriv[, ii]) * (Ws.intF.vl.alph * Xs[, 
                                                    j] + Ws.intF.sl.alph * Xs.deriv[, jj])
        }
        else if (i %in% indFixed && !j %in% indFixed) {
          ii <- match(i, indFixed)
          (Ws.intF.vl.alph * Xs[, i] + Ws.intF.sl.alph * 
            Xs.deriv[, ii]) * (Ws.intF.vl.alph * Xs[, 
                                                    j])
        }
        else if (!i %in% indFixed && j %in% indFixed) {
          jj <- match(j, indFixed)
          (Ws.intF.vl.alph * Xs[, i]) * (Ws.intF.vl.alph * 
                                           Xs[, j] + Ws.intF.sl.alph * Xs.deriv[, jj])
        }
        else {
          Ws.intF.vl.alph^2 * Xs[, i] * Xs[, j]
        }
      }
      ki <- exp.eta.tw.P * rowsum(Int * XX, id.GK, reorder = FALSE)
      kii <- c((p.byt * ki) %*% wGH)
      H2[i, j] <- sum(kii, na.rm = TRUE)
    }
  }
  H2[lower.tri(H2)] <- t(H2)[lower.tri(H2)]
  H1 + H2
}

LogLik.weibullGH = function (thetas) 
{
  thetas <- relist(thetas, skeleton = list.thetas)
  betas <- thetas$betas
  sigma <- exp(thetas$log.sigma)
  gammas <- thetas$gammas
  alpha <- thetas$alpha
  Dalpha <- thetas$Dalpha
  sigma.t <- if (is.null(scaleWB)) 
    exp(thetas$log.sigma.t)
  else scaleWB
  D <- thetas$D
  D <- if (diag.D) 
    exp(D)
  else chol.transf(D)
  eta.yx <- as.vector(X %*% betas)
  eta.tw <- as.vector(WW %*% gammas)
  if (parameterization %in% c("value", "both")) {
    Y <- as.vector(Xtime %*% betas) + Ztime.b
    Ys <- as.vector(Xs %*% betas) + Zsb
    eta.t <- eta.tw + c(WintF.vl %*% alpha) * Y
    eta.s <- c(Ws.intF.vl %*% alpha) * Ys
  }
  if (parameterization %in% c("slope", "both")) {
    Y.deriv <- as.vector(Xtime.deriv %*% betas[indFixed]) + 
      Ztime.b.deriv
    Ys.deriv <- as.vector(Xs.deriv %*% betas[indFixed]) + 
      Zsb.deriv
    eta.t <- if (parameterization == "both") 
      eta.t + c(WintF.sl %*% Dalpha) * Y.deriv
    else eta.tw + c(WintF.sl %*% Dalpha) * Y.deriv
    eta.s <- if (parameterization == "both") 
      eta.s + c(Ws.intF.sl %*% Dalpha) * Ys.deriv
    else c(Ws.intF.sl %*% Dalpha) * Ys.deriv
  }
  mu.y <- eta.yx + Ztb
  logNorm <- dnorm(y, mu.y, sigma, TRUE)
  log.p.yb <- rowsum(logNorm, id)
  log.hazard <- log(sigma.t) + (sigma.t - 1) * logT + eta.t
  log.survival <- -exp(eta.tw) * P * rowsum(wk * exp(log(sigma.t) + 
                                                       (sigma.t - 1) * log.st + eta.s), id.GK, reorder = FALSE)
  dimnames(log.survival) <- NULL
  log.p.tb <- d * log.hazard + log.survival
  log.p.b <- if (control$typeGH == "simple") {
    rep(dmvnorm(b, rep(0, ncz), D, TRUE), each = n)
  }
  else {
    matrix(dmvnorm(do.call(rbind, lis.b), rep(0, ncz), D, 
                   TRUE), n, k, byrow = TRUE)
  }
  p.ytb <- exp(log.p.yb + log.p.tb + log.p.b)
  if (control$typeGH != "simple") 
    p.ytb <- p.ytb * VCdets
  dimnames(p.ytb) <- NULL
  p.yt <- c(p.ytb %*% wGH)
  log.p.yt <- log(p.yt)
  -sum(log.p.yt[is.finite(log.p.yt)], na.rm = TRUE)
}

Score.weibullGH = function (thetas) 
{
  thetas <- relist(thetas, skeleton = list.thetas)
  betas <- thetas$betas
  sigma <- exp(thetas$log.sigma)
  gammas <- thetas$gammas
  alpha <- thetas$alpha
  Dalpha <- thetas$Dalpha
  sigma.t <- if (is.null(scaleWB)) 
    exp(thetas$log.sigma.t)
  else scaleWB
  D <- thetas$D
  D <- if (diag.D) 
    exp(D)
  else chol.transf(D)
  eta.yx <- as.vector(X %*% betas)
  eta.tw <- as.vector(WW %*% gammas)
  if (parameterization %in% c("value", "both")) {
    Y <- as.vector(Xtime %*% betas) + Ztime.b
    Ys <- as.vector(Xs %*% betas) + Zsb
    WintF.vl.alph <- c(WintF.vl %*% alpha)
    Ws.intF.vl.alph <- c(Ws.intF.vl %*% alpha)
    eta.t <- eta.tw + WintF.vl.alph * Y
    eta.s <- Ws.intF.vl.alph * Ys
  }
  if (parameterization %in% c("slope", "both")) {
    Y.deriv <- as.vector(Xtime.deriv %*% betas[indFixed]) + 
      Ztime.b.deriv
    Ys.deriv <- as.vector(Xs.deriv %*% betas[indFixed]) + 
      Zsb.deriv
    WintF.sl.alph <- c(WintF.sl %*% Dalpha)
    Ws.intF.sl.alph <- c(Ws.intF.sl %*% Dalpha)
    eta.t <- if (parameterization == "both") 
      eta.t + WintF.sl.alph * Y.deriv
    else eta.tw + WintF.sl.alph * Y.deriv
    eta.s <- if (parameterization == "both") 
      eta.s + Ws.intF.sl.alph * Ys.deriv
    else Ws.intF.sl.alph * Ys.deriv
  }
  exp.eta.tw <- exp(eta.tw)
  exp.eta.tw.P <- exp.eta.tw * P
  mu.y <- eta.yx + Ztb
  logNorm <- dnorm(y, mu.y, sigma, TRUE)
  log.p.yb <- rowsum(logNorm, id)
  log.hazard <- log(sigma.t) + (sigma.t - 1) * logT + eta.t
  Int <- wk * exp(log(sigma.t) + (sigma.t - 1) * log.st + eta.s)
  log.survival <- -exp.eta.tw * P * rowsum(Int, id.GK, reorder = FALSE)
  dimnames(log.survival) <- NULL
  log.p.tb <- d * log.hazard + log.survival
  log.p.b <- if (control$typeGH == "simple") {
    rep(dmvnorm(b, rep(0, ncz), D, TRUE), each = n)
  }
  else {
    matrix(dmvnorm(do.call(rbind, lis.b), rep(0, ncz), D, 
                   TRUE), n, k, byrow = TRUE)
  }
  p.ytb <- exp(log.p.yb + log.p.tb + log.p.b)
  if (control$typeGH != "simple") 
    p.ytb <- p.ytb * VCdets
  dimnames(p.ytb) <- NULL
  p.yt <- c(p.ytb %*% wGH)
  p.byt <- p.ytb/p.yt
  post.b <- if (control$typeGH == "simple") {
    p.byt %*% (b * wGH)
  }
  else {
    sapply(seq_len(ncz), function(i) (p.byt * t(sapply(lis.b, 
                                                       "[", seq_len(k), i))) %*% wGH)
  }
  post.vb <- if (control$typeGH == "simple") {
    if (ncz == 1) {
      c(p.byt %*% (b2 * wGH)) - c(post.b * post.b)
    }
    else {
      (p.byt %*% (b2 * wGH)) - t(apply(post.b, 1, function(x) x %o% 
                                         x))
    }
  }
  else {
    dd <- sapply(seq_len(ncz^2), function(i) (p.byt * t(sapply(lis.b2, 
                                                               "[", seq_len(k), i))) %*% wGH)
    bb <- apply(post.b, 1, function(x) x %o% x)
    dd - if (ncz == 1) 
      c(bb)
    else t(bb)
  }
  Zb <- if (ncz == 1) 
    post.b[id]
  else rowSums(Z * post.b[id, ], na.rm = TRUE)
  mu <- y - eta.yx
  tr.tZZvarb <- sum(ZtZ * post.vb, na.rm = TRUE)
  sc1 <- -crossprod(X, y - eta.yx - Zb)/sigma^2
  exp.eta.tw.P <- exp.eta.tw * P
  sc2 <- numeric(ncx)
  for (i in 1:ncx) {
    ki <- exp.eta.tw.P * switch(parameterization, value = rowsum(Int * 
                                                                   Ws.intF.vl.alph * Xs[, i], id.GK, reorder = FALSE), 
                                slope = {
                                  ii <- match(i, indFixed)
                                  if (is.na(ii)) 0 else rowsum(Int * Ws.intF.sl.alph * 
                                                                 Xs.deriv[, ii], id.GK, reorder = FALSE)
                                }, both = {
                                  ii <- match(i, indFixed)
                                  rowsum(Int * (Ws.intF.vl.alph * Xs[, i] + Ws.intF.sl.alph * 
                                                  if (is.na(ii)) 0 else Xs.deriv[, ii]), id.GK, 
                                         reorder = FALSE)
                                })
    kii <- c((p.byt * ki) %*% wGH)
    sc2[i] <- switch(parameterization, value = -sum(d * WintF.vl.alph * 
                                                      Xtime[, i] - kii, na.rm = TRUE), slope = {
                                                        ii <- match(i, indFixed)
                                                        if (is.na(ii)) 0 else -sum(d * WintF.sl.alph * Xtime.deriv[, 
                                                                                                                   ii] - kii, na.rm = TRUE)
                                                      }, both = {
                                                        ii <- match(i, indFixed)
                                                        -sum(d * (WintF.vl.alph * Xtime[, i] + WintF.sl.alph * 
                                                                    if (is.na(ii)) 0 else Xtime.deriv[, ii]) - kii, 
                                                             na.rm = TRUE)
                                                      })
  }
  score.y <- c(sc1 + sc2, -sigma * (-N/sigma + drop(crossprod(mu, 
                                                              mu - 2 * Zb) + crossprod(Zb) + tr.tZZvarb)/sigma^3))
  ki <- P * rowsum(Int, id.GK, reorder = FALSE)
  kii <- c((p.byt * ki) %*% wGH)
  scgammas <- -colSums(WW * (d - exp.eta.tw * kii), na.rm = TRUE)
  scalpha <- if (parameterization %in% c("value", "both")) {
    rr <- numeric(ncol(WintF.vl))
    for (l in seq_along(rr)) rr[l] <- -sum((p.byt * (d * 
                                                       WintF.vl[, l] * Y - exp.eta.tw.P * rowsum(Int * Ws.intF.vl[, 
                                                                                                                  l] * Ys, id.GK, reorder = FALSE))) %*% wGH, na.rm = TRUE)
    rr
  }
  else NULL
  scalpha.D <- if (parameterization %in% c("slope", "both")) {
    rr <- numeric(ncol(WintF.sl))
    for (l in seq_along(rr)) rr[l] <- -sum((p.byt * (d * 
                                                       WintF.sl[, l] * Y.deriv - exp.eta.tw.P * rowsum(Int * 
                                                                                                         Ws.intF.sl[, l] * Ys.deriv, id.GK, reorder = FALSE))) %*% 
                                             wGH, na.rm = TRUE)
    rr
  }
  else NULL
  scsigmat <- if (is.null(scaleWB)) {
    Int2 <- st^(sigma.t - 1) * (1 + sigma.t * log.st) * exp(eta.s)
    -sigma.t * sum((p.byt * (d * (1/sigma.t + logT) - exp.eta.tw.P * 
                               rowsum(wk * Int2, id.GK, reorder = FALSE))) %*% wGH, 
                   na.rm = TRUE)
  }
  else NULL
  score.t <- c(scgammas, scalpha, scalpha.D, scsigmat)
  score.b <- if (diag.D) {
    svD <- 1/D
    svD2 <- svD^2
    cS.postVB <- colSums(as.matrix(post.vb), na.rm = TRUE)
    dim(cS.postVB) <- c(ncz, ncz)
    D * 0.5 * (n * svD - diag(cS.postVB) * svD2 - colSums(as.matrix(post.b^2), 
                                                          na.rm = TRUE) * svD2)
  }
  else {
    svD <- solve(D)
    dD <- deriv.D(D)
    ndD <- length(dD)
    D1 <- sapply(dD, function(x) sum(svD * x))
    D2 <- t(sapply(dD, function(x) c(svD %*% x %*% svD)))
    cS.postVB <- colSums(as.matrix(post.vb), na.rm = TRUE)
    out <- numeric(ndD)
    for (i in seq_along(dD)) {
      D.mat <- D2[i, ]
      dim(D.mat) <- c(ncz, ncz)
      out[i] <- sum(D2[i, ] * cS.postVB, na.rm = TRUE) + 
        sum((post.b %*% D.mat) * post.b, na.rm = TRUE)
    }
    J <- jacobian2(attr(D, "L"), ncz)
    drop(0.5 * (n * D1 - out) %*% J)
  }
  c(score.y, score.t, score.b)
}

ModelMats = function (time, ii, obs.times, survTimes) 
{
  if (method %in% c("weibull-PH-GH")) {
    id.GK <- if (!LongFormat) {
      rep(ii, each = object$control$GKk)
    }
    else {
      rep(which(idT == ii), each = object$control$GKk)
    }
    wk <- gaussKronrod(object$control$GKk)$wk
    sk <- gaussKronrod(object$control$GKk)$sk
    if (!LongFormat) {
      P <- time/2
      st <- P * (sk + 1)
    }
    else {
      time0 <- obs.times[[ii]]
      time1 <- c(time0[-1], time)
      P <- (time1 - time0)/2
      P1 <- (time1 + time0)/2
      st <- outer(P, sk) + P1
      st <- c(t(st))
    }
    data.id2 <- data.id[id.GK, ]
    data.id2[[timeVar]] <- pmax(st - lag, 0)
    out <- list(st = st, wk = rep(wk, length(P)), P = P)
    if (parameterization %in% c("value", "both")) {
      mfX <- model.frame(delete.response(TermsX), data = data.id2)
      mfZ <- model.frame(TermsZ, data = data.id2)
      out$Xs <- model.matrix(formYx, mfX)
      out$Zs <- model.matrix(formYz, mfZ)
      out$Ws.intF.vl <- WintF.vl[id.GK, , drop = FALSE]
    }
    if (parameterization %in% c("slope", "both")) {
      mfX.deriv <- model.frame(TermsX.deriv, data = data.id2)
      mfZ.deriv <- model.frame(TermsZ.deriv, data = data.id2)
      out$Xs.deriv <- model.matrix(derivForm$fixed, mfX.deriv)
      out$Zs.deriv <- model.matrix(derivForm$random, mfZ.deriv)
      out$Ws.intF.sl <- WintF.sl[id.GK, , drop = FALSE]
    }
  }
  if (method == "piecewise-PH-GH") {
    wk <- gaussKronrod(7)$wk
    sk <- gaussKronrod(7)$sk
    nk <- length(sk)
    qs <- c(0, sort(object$control$knots), max(survTimes, 
                                               object$control$knots) + 1)
    ind <- findInterval(time, qs, rightmost.closed = TRUE)
    Tiq <- outer(time, qs, pmin)
    Lo <- Tiq[, 1:Q]
    Up <- Tiq[, 2:(Q + 1)]
    T <- Up - Lo
    P <- T/2
    if (!all(P < sqrt(.Machine$double.eps))) 
      P[P < sqrt(.Machine$double.eps)] <- as.numeric(NA)
    P1 <- (Up + Lo)/2
    st <- rep(P, each = nk) * rep(sk, Q) + rep(P1, each = nk)
    data.id2 <- data.id[rep(ii, each = nk * Q), ]
    data.id2[[timeVar]] <- pmax(st - lag, 0)
    data.id2 <- data.id2[!is.na(st), ]
    id.GK <- rep(ii, sum(!is.na(st)))
    out <- list(st = st, wk = wk, P = P, ind = ind)
    if (parameterization %in% c("value")) {
      mfX <- model.frame(TermsX, data = data.id2)
      mfZ <- model.frame(TermsZ, data = data.id2)
      out$Xs <- model.matrix(formYx, mfX)
      out$Zs <- model.matrix(object$formYz, mfZ)
      out$Ws.intF.vl <- WintF.vl[id.GK, , drop = FALSE]
    }
  }
  out
}

cd = function (x, f, ..., eps = 0.001) 
{
  n <- length(x)
  res <- numeric(n)
  ex <- pmax(abs(x), 1)
  for (i in 1:n) {
    x1 <- x2 <- x
    x1[i] <- x[i] + eps * ex[i]
    x2[i] <- x[i] - eps * ex[i]
    diff.f <- c(f(x1, ...) - f(x2, ...))
    diff.x <- x1[i] - x2[i]
    res[i] <- diff.f/diff.x
  }
  res
}

rmvt = function (n, mu, Sigma, df) 
{
  mvrnorm(n, mu = mu, Sigma = Sigma)/rep(sqrt(rchisq(n, df)/df), 
                                         length(mu))
}

dmvt = function (x, mu, Sigma, df, log = FALSE) 
{
  if (!is.numeric(x)) 
    stop("'x' must be a numeric matrix or vector")
  if (!is.matrix(x)) 
    x <- rbind(x)
  p <- length(mu)
  if (!all(dim(Sigma) == c(p, p)) || ncol(x) != p) 
    stop("incompatible arguments")
  ed <- eigen(Sigma, symmetric = TRUE)
  ev <- ed$values
  if (!all(ev >= -1e-06 * abs(ev[1]))) 
    stop("'Sigma' is not positive definite")
  ss <- x - rep(mu, each = nrow(x))
  inv.Sigma <- ed$vectors %*% (t(ed$vectors)/ev)
  quad <- rowSums((ss %*% inv.Sigma) * ss)/df
  fact <- lgamma((df + p)/2) - lgamma(df/2) - 0.5 * (p * (log(pi) + 
                                                            log(df)) + sum(log(ev)))
  if (log) 
    fact - 0.5 * (df + p) * log(1 + quad)
  else exp(fact) * ((1 + quad)^(-(df + p)/2))
}

